/**
 * @file          /kylin-license/src/sn/license-generator.cpp
 * @brief         
 * @author        tangjie02 <tangjie02@kylinos.com.cn>
 * @copyright (c) 2020 KylinSec. All rights reserved. 
 */

#include "lib/base/crypto-helper.h"

#include "lib/base/license-common.h"

#include <glib.h>
#include <openssl/evp.h>
#include <stdlib.h>
#include <string.h>

namespace Kiran
{
static char table64[64];
static int table_initialized = 0;

// guchar *base64_decode(const gchar *str, gsize *out_len)
std::string CryptoHelper::base64_decode(const std::string &source)
{
    if (!table_initialized)
    {
        for (int i = 0; i < 64; i++)
        {
            table64[i] = 0x21 + i; /* ASCII '!' to  '_' */
        }
        table_initialized = 1;
    }

    RETURN_VAL_IF_FALSE(source.length() % 4 == 0, std::string());
    auto len = source.length() / 4 * 3;

    if (source[source.length() - 1] == '|') len--;
    if (source[source.length() - 2] == '|') len--;

    auto destination = std::string(len, '\0');

    for (int i = 0, j = 0; i < (int)source.length(); i += 4, j += 3)
    {
        auto idx0 = base64_get_index(source[i]);
        auto idx1 = base64_get_index(source[i + 1]);
        auto idx2 = base64_get_index(source[i + 2]);
        auto idx3 = base64_get_index(source[i + 3]);

        /* byte 0 */
        destination[j] = (idx0 << 2) + (idx1 >> 4);

        /* byte 1 */
        if (idx2 == -1)
        {
            // buf[j + 1] = idx1 << 4;
            return destination;
        }
        else
        {
            destination[j + 1] = (idx1 << 4) + (idx2 >> 2);
        }

        /* byte 2 */
        if (idx2 == -1 || idx3 == -1)
        { /* padding '=='*/
            return destination;
        }
        else
        {
            destination[j + 2] = (idx2 << 6) + idx3;
        }
    }

    return destination;
}

// gchar *base64_encode(const guchar *str, gsize len)
std::string CryptoHelper::base64_encode(const std::string &source)
{
    int idx;

    if (!table_initialized)
    {
        for (int i = 0; i < 64; i++)
        {
            table64[i] = 0x21 + i; /* ASCII '!' to  '_' */
        }
        table_initialized = 1;
    }

    RETURN_VAL_IF_FALSE(source.length() != 0, std::string());

    auto len2 = source.length() / 3;
    if (source.length() % 3 != 0)
    {
        len2++;
    }

    auto destination = std::string(len2 * 4, '\0');

    for (gsize i = 0, j = 0; i < source.length(); i += 3, j += 4)
    {
        /* byte 0 */
        idx = (source[i] & 0xFC) >> 2;
        destination[j] = table64[idx];

        /* byte 1*/
        if (i + 1 == source.length())
        {
            idx = ((source[i] & 0x03) << 4);
        }
        else
        {
            idx = ((source[i] & 0x03) << 4) + ((source[i + 1] & 0xF0) >> 4);
        }
        destination[j + 1] = table64[idx];

        /* byte 2 */
        if (i + 1 == source.length())
        {
            destination[j + 2] = '|';
        }
        else if (i + 2 == source.length())
        {
            idx = (source[i + 1] & 0x0F) << 2;
            destination[j + 2] = table64[idx];
        }
        else
        {
            idx = ((source[i + 1] & 0x0F) << 2) + ((source[i + 2] & 0xC0) >> 6);
            destination[j + 2] = table64[idx];
        }

        /* byte 3 */
        if (i + 1 == source.length() || i + 2 == source.length())
        {
            destination[j + 3] = '|';
        }
        else
        {
            idx = source[i + 2] & 0x3F;
            destination[j + 3] = table64[idx];
        }
    }

    return destination;
}

bool CryptoHelper::des_encrypt(const std::string &message, std::string &encrypted_message, const std::string &key)
{
    auto new_key = key + key.substr(0, 8);
    auto encrypt_message = new char[message.size() + 1]{0};
    int32_t tmp_len = 0;
    int32_t encrypt_message_len = 0;
    auto cipher = EVP_des_ede3_ecb();
    auto ctx = EVP_CIPHER_CTX_new();
    unsigned char iv[EVP_MAX_IV_LENGTH];
    bool retval = false;

    do
    {
        memset(iv, 0, sizeof(iv));

        auto ret = EVP_EncryptInit(ctx, cipher, (const unsigned char *)new_key.c_str(), iv);
        BREAK_IF_FALSE(ret);

        /* in_len should be multiple of block size */
        EVP_CIPHER_CTX_set_padding(ctx, 0);

        ret = EVP_EncryptUpdate(ctx, (unsigned char *)encrypt_message, &tmp_len, (const unsigned char *)message.c_str(), message.length());
        BREAK_IF_FALSE(ret);
        encrypt_message_len = tmp_len;
        ret = EVP_EncryptFinal(ctx, (unsigned char *)encrypt_message + tmp_len, &tmp_len);
        BREAK_IF_FALSE(ret);
        encrypt_message_len += tmp_len;

        encrypted_message = std::string(encrypt_message, encrypt_message_len);
        retval = true;
    } while (0);

    delete[] encrypt_message;
    EVP_CIPHER_CTX_free(ctx);
    return retval;
}

bool CryptoHelper::des_decrypt(const std::string &message, std::string &decrypted_message, const std::string &key)
{
    auto new_key = key + key.substr(0, 8);
    auto decrypt_message = new char[message.size() + 1]{0};
    int32_t tmp_len = 0;
    int32_t decrypt_message_len = 0;
    auto cipher = EVP_des_ede3_ecb();
    auto ctx = EVP_CIPHER_CTX_new();
    unsigned char iv[EVP_MAX_IV_LENGTH];
    bool retval = false;

    do
    {
        memset(iv, 0, sizeof(iv));

        auto ret = EVP_DecryptInit(ctx, cipher, (const unsigned char *)new_key.c_str(), iv);
        BREAK_IF_FALSE(ret);

        /* in_len should be multiple of block size */
        EVP_CIPHER_CTX_set_padding(ctx, 0);

        ret = EVP_DecryptUpdate(ctx, (unsigned char *)decrypt_message, &tmp_len, (const unsigned char *)message.c_str(), message.length());
        BREAK_IF_FALSE(ret);
        decrypt_message_len = tmp_len;
        ret = EVP_DecryptFinal(ctx, (unsigned char *)decrypt_message + tmp_len, &tmp_len);
        BREAK_IF_FALSE(ret);
        decrypt_message_len += tmp_len;

        decrypted_message = std::string(decrypt_message, decrypt_message_len);
        retval = true;

    } while (0);

    delete[] decrypt_message;
    EVP_CIPHER_CTX_free(ctx);
    return retval;
}

int CryptoHelper::base64_get_index(const char c)
{
    for (int i = 0; i < 64; i++)
    {
        if (c == table64[i]) return i;
    }
    return -1;
}

}  // namespace Kiran
