/**
 * @file          /kylin-license/lib/base/file-utils.cpp
 * @brief         
 * @author        tangjie02 <tangjie02@kylinos.com.cn>
 * @copyright (c) 2020 KylinSec. All rights reserved. 
 */

#include "lib/base/file-utils.h"
#include <fcntl.h>
#include <gtk3-log-i.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <unistd.h>
#include "lib/base/license-common.h"

namespace Kiran
{
Glib::RefPtr<Gio::FileMonitor> FileUtils::make_monitor_file(const std::string &path,
                                                            const FileMonitorCallBack &callback,
                                                            Gio::FileMonitorFlags flags)
{
    auto file = Gio::File::create_for_path(path);
    try
    {
        auto monitor = file->monitor_file(flags);
        monitor->signal_changed().connect(callback);
        return monitor;
    }
    catch (const Glib::Error &e)
    {
        KLOG_WARNING("Unable to file monitor %s: %s", path.c_str(), e.what().c_str());
    }

    return Glib::RefPtr<Gio::FileMonitor>();
}

std::string FileUtils::read_bdata(const std::string &file_path)
{
    std::string retval;
    int fp = -1;
    gchar *buf = NULL;

    do
    {
        fp = open(file_path.c_str(), O_RDONLY);
        if (fp < 0)
        {
            KLOG_WARNING("Failed to open file %s: %s.", file_path.c_str(), strerror(errno));
            break;
        }
        auto file_size = FileUtils::get_file_size(fp);
        BREAK_IF_TRUE(file_size < 0);

        buf = (char *)g_malloc0(file_size);
        if (read(fp, buf, file_size) < 0)
        {
            KLOG_WARNING("Failed to read file %s: %s.", file_path.c_str(), strerror(errno));
            break;
        }
        retval = std::string(buf, file_size);
    } while (0);

    if (fp > 0)
    {
        close(fp);
    }

    if (buf)
    {
        g_free(buf);
    }
    return retval;
}

bool FileUtils::write_bdata(const std::string &file_path, const std::string &data, int oflag, int mode_flag)
{
    KLOG_DEBUG("File path: %s", file_path.c_str());

    int fp = -1;
    bool retval = false;

    do
    {
        fp = open(file_path.c_str(), oflag, mode_flag);

        if (fp < 0)
        {
            KLOG_WARNING("Failed to open file %s: %s.", file_path.c_str(), strerror(errno));
            break;
        }

        if (write(fp, data.c_str(), data.length()) < 0)
        {
            KLOG_WARNING("Failed to write file %s: %s.", file_path.c_str(), strerror(errno));
            break;
        }
        retval = true;
    } while (0);

    if (fp > 0)
    {
        close(fp);
    }
    return retval;
}

int FileUtils::get_file_size(int fd)
{
    struct stat stmp;
    if (fstat(fd, &stmp) < 0)
    {
        KLOG_WARNING("Failed to get file size: %s.", strerror(errno));
        return -1;
    }
    return stmp.st_size;
}
}  // namespace Kiran
