/**
 * @file          /kylin-license/lib/base/license-utils.cpp
 * @brief         
 * @author        tangjie02 <tangjie02@kylinos.com.cn>
 * @copyright (c) 2020 KylinSec. All rights reserved. 
 */
#include "lib/base/license-utils.h"

#include <gtk3-log-i.h>
#include "lib/base/crypto-helper.h"
#include "lib/base/license-common.h"
#include "lib/base/time-utils.h"

namespace Kiran
{
#define VIRT_WHAT_CMD "/usr/sbin/virt-what"

#define LICENSE_CLIENT_ENCODE_KEY "KyLINO\343'S SECRET"

// 该KEY是用来通过一个字符串来生成另一个KEY的，另一个KEY用于算号程序
#define KEY_LICENSE_KEY "key__license_key"

bool LicenseUtils::is_ksvd()
{
    FILE *fp;
    char *line = NULL;
    size_t len;
    bool ret = false;

    fp = fopen("/proc/cpuinfo", "r");
    if (!fp) return ret;

    while (getline(&line, &len, fp) != -1)
    {
        if (strstr(line, "model name"))
        {
            if (strstr(line, "KSVD Virtual CPU"))
            {
                ret = true;
            }
            break;
        }
    }

    if (line) free(line);
    fclose(fp);

    return ret;
}

bool LicenseUtils::is_virtual_machine()
{
    std::string standard_output;
    std::string standard_error;
    gint exit_status = 0;

    std::vector<std::string> argv{VIRT_WHAT_CMD};

    try
    {
        Glib::spawn_sync(std::string(), argv, Glib::SpawnFlags(0), sigc::slot<void>(), &standard_output, &standard_error, &exit_status);
    }
    catch (const Glib::Error &e)
    {
        KLOG_WARNING("Command %s error: %s exit status: %d. g_spawn_sync error: %s.",
                     VIRT_WHAT_CMD,
                     standard_error.c_str(),
                     exit_status,
                     e.what().c_str());
        return false;
    }

    KLOG_DEBUG("virtual machine: %s.", standard_output.c_str());
    return standard_output.size() > 0;
}

bool LicenseUtils::is_localhost(const std::string &host)
{
    if (host.find("127.0.0.1") != std::string::npos ||
        host.find("localhost") != std::string::npos)
    {
        return true;
    }
    return false;
}

std::string LicenseUtils::binary_to_code(const unsigned char *in_binary, int binary_len)
{
    std::string result;
    char buf[3];

    RETURN_VAL_IF_TRUE(in_binary == NULL, NULL);
    RETURN_VAL_IF_TRUE(binary_len <= 0, NULL);

    for (int i = 0, j = 0; i < binary_len; i++, j += 2)
    {
        if (i != 0 && i % 2 == 0)
        {
            result.push_back('-');
            j++;
        }
        g_snprintf(buf, 3, "%02X", in_binary[i]);
        result.append(buf);
    }

    return result;
}

// code format: xxxx-xxxx-xxxx
bool LicenseUtils::code_to_binary(const char *in_str, unsigned char out_binary[], int out_len)
{
    int pos = 0;
    char buf[3] = {0};
    int fill_len = 0;

    RETURN_VAL_IF_TRUE(in_str == NULL, false);

    while (in_str[pos] != '\0' && fill_len < out_len)
    {
        if (!isxdigit(in_str[pos]))
        {
            pos += 1;
            continue;
        }
        if (!isxdigit(in_str[pos + 1]))
        {
            return false;
        }

        memcpy(buf, in_str + pos, 2);
        out_binary[fill_len++] = (unsigned char)strtoul(buf, NULL, 16);
        pos += 2;
    }

    return (in_str[pos] == '\0' && fill_len == out_len);
}

static int char2digit(char c)
{
    if (c >= '0' && c <= '9')
    {
        return c - '0';
    }
    else if (c >= 'A' && c <= 'Z')
    {
        return 10 + c - 'A';
    }
    KLOG_WARNING("Character out of bounds: %c.", c);
    return 0;
}

static char digit2char(int digit)
{
    if (digit >= 0 && digit < 10)
    {
        return '0' + digit;
    }
    else if (digit >= 10 && digit < 36)
    {
        return 'A' + digit - 10;
    }
    KLOG_WARNING("digit out of bounds: %d.", digit);
    return '0';
}

static std::string division(std::string in_str,
                            int dst_base,
                            int org_base,
                            int *remain)
{
    int a = 0;
    *remain = 0;

    for (int i = 0; i < (int)in_str.size(); i++)
    {
        a = (org_base * (*remain) + char2digit(in_str[i]));
        in_str[i] = digit2char(a / dst_base);
        *remain = a % dst_base;
    }
    //去掉多余的0 比如10/2=05
    int pos = 0;
    while (in_str[pos] == '0' && pos < (int)in_str.size())
    {
        pos++;
    }
    return in_str.substr(pos);
}

// 大数进制转换
static char *base_conversion(const char *in_str,
                             int in_str_len,
                             int out_str_len,
                             int dst_base,
                             int org_base)
{
    int remain = 0;
    GString *out_str = NULL;
    gchar *result = NULL;

    out_str = g_string_new(NULL);
    auto str1 = std::string(in_str, in_str_len);

    //因为去掉了多余的0，所以终止条件是字符串为空
    //例：当上一步运算结果为"0"时，实际上返回的结果为""
    while (str1.size() != 0)
    {
        str1 = division(str1, dst_base, org_base, &remain);
        out_str = g_string_append_c(out_str, digit2char(remain));
    }
    // 位数不足用'0'填充
    while (out_str->len < (gsize)out_str_len)
    {
        g_string_append_c(out_str, '0');
    }
    result = g_string_free(out_str, FALSE);

    return g_strreverse(result);
}

static void bin2bit(const unsigned char *bin, int bin_len, char *bit, int *bit_len)
{
    int out_len = 0;

    for (int i = 0; i < bin_len; ++i)
    {
        for (int j = 7; j >= 0 && out_len < *bit_len; --j)
        {
            if ((bin[i] & (1 << j)) > 0)
            {
                bit[out_len++] = '1';
            }
            else
            {
                bit[out_len++] = '0';
            }
        }
    }
    *bit_len = out_len;
}

static bool bit2bin(char *bit, unsigned char *bin_str, int *bin_str_len)
{
    int out_len = 0;
    GString *bit_str = NULL;
    int i = 0;
    int j = 0;

    if (strlen(bit) != LICENSE_CODE_BIT_LENGTH)
    {
        KLOG_WARNING("Invalid length: %d", (int)strlen(bit));
        return false;
    }

    bit_str = g_string_new(bit);

    for (i = 0; i < (int)bit_str->len && out_len < *bin_str_len; i += 8)
    {
        char c = 0;
        for (j = i; j < i + 8; ++j)
        {
            if (bit_str->str[j] != '0' && bit_str->str[j] != '1')
            {
                KLOG_WARNING("Invalid bit str");
                return false;
            }
            c = c * 2 + bit_str->str[j] - '0';
        }
        bin_str[out_len++] = c;
    }
    *bin_str_len = out_len;
    return true;
}

static std::string format_license_show_code(char *in_str)
{
    int in_str_len = strlen(in_str);
    int i = 0;

    auto out_str = std::string(in_str);
    // xxxxx-xxxxx-xxxxx-xxxxx-xxxxx
    if (in_str_len == LICENSE_CODE_SHOW_LENGHT)
    {
        for (i = LICENSE_CODE_SHOW_LENGHT - 5; i > 0; i -= 5)
        {
            out_str.insert(i, 1, '-');
        }
    }
    else if (in_str_len == LICENSE_CODE_LONG_SHOW_LENGHT)
    {
        for (i = LICENSE_CODE_LONG_SHOW_LENGHT - 4; i > 0; i -= 4)
        {
            out_str.insert(i, 1, '-');
        }
    }
    else
    {
        KLOG_WARNING("Invalid length: %d", in_str_len);
        return NULL;
    }
    return out_str;
}

std::string LicenseUtils::unformat_license_show_code(std::string str)
{
    std::string retval;
    for (uint32_t i = 0; i < str.size(); ++i)
    {
        if (str[i] != '-')
        {
            retval.push_back(str[i]);
        }
    }
    return retval;
}

std::string LicenseUtils::format_activation_code(const std::string &activation_code)
{
    char license_code_bit[LICENSE_CODE_BIT_LENGTH];
    int bit_len = LICENSE_CODE_BIT_LENGTH;
    bin2bit((const unsigned char *)activation_code.c_str(), LICENSE_CODE_LENGTH, license_code_bit, &bit_len);
    auto license_code_str = base_conversion((char *)license_code_bit, bit_len, LICENSE_CODE_SHOW_LENGHT, LICENSE_CODE_SHOW_BASE, 2);
    auto retval = format_license_show_code(license_code_str);
    g_free(license_code_str);
    return retval;
}

std::string LicenseUtils::format_long_activation_code(const std::string &activation_code)
{
    char license_code_bit[LICENSE_CODE_BIT_LENGTH];
    int bit_len = LICENSE_CODE_BIT_LENGTH;
    bin2bit((const unsigned char *)activation_code.c_str(), LICENSE_CODE_LENGTH, license_code_bit, &bit_len);
    auto license_code_str = base_conversion((char *)license_code_bit, bit_len, LICENSE_CODE_LONG_SHOW_LENGHT, LICENSE_CODE_LONG_SHOW_BASE, 2);
    auto retval = format_license_show_code(license_code_str);
    g_free(license_code_str);
    return retval;
}

LicenseErrorCode LicenseUtils::unformat_activation_code(const std::string &str, std::string &activation_code)
{
    int license_code_len = LICENSE_CODE_LENGTH;
    auto unformat_str = unformat_license_show_code(str);
    activation_code.resize(LICENSE_CODE_LENGTH);
    // 这里兼容以前的激活码格式
    if (unformat_str.length() == LICENSE_CODE_LONG_SHOW_LENGHT)
    {
        if (!code_to_binary(str.c_str(), (unsigned char *)activation_code.data(), LICENSE_CODE_LENGTH))
        {
            return LicenseErrorCode::LEC_LICENSE_TRANS_FAILED;
        }
    }
    else if (unformat_str.length() == LICENSE_CODE_SHOW_LENGHT)
    {
        auto license_code_bit = base_conversion(unformat_str.c_str(), unformat_str.length(), LICENSE_CODE_BIT_LENGTH, 2, LICENSE_CODE_SHOW_BASE);
        if (!bit2bin(license_code_bit, (unsigned char *)activation_code.data(), &license_code_len))
        {
            g_free(license_code_bit);
            return LicenseErrorCode::LEC_LICENSE_TRANS_FAILED;
        }
        g_free(license_code_bit);
    }
    else
    {
        return LicenseErrorCode::LEC_LICENSE_LENGTH_INVALID;
    }
    return LicenseErrorCode::LEC_SUCCESS;
}

std::string LicenseUtils::format_machine_code(const std::string &machine_code)
{
    return LicenseUtils::binary_to_code((const unsigned char *)machine_code.data(), MACHINE_CODE_LENGTH);
}

LicenseErrorCode LicenseUtils::unformat_machine_code(const std::string &str, std::string &machine_code)
{
    machine_code.resize(MACHINE_CODE_LENGTH);
    if (!LicenseUtils::code_to_binary(str.c_str(), (unsigned char *)machine_code.data(), MACHINE_CODE_LENGTH))
    {
        return LicenseErrorCode::LEC_MACHINE_CODE_TRANS_FAILED;
    }
    return LicenseErrorCode::LEC_SUCCESS;
}

std::string LicenseUtils::generate_key(const std::string &name)
{
    // 这里需要兼容旧的版本
    if (name == LICENSE_OBJECT_OS_NAME || name.empty())
    {
        return LICENSE_CLIENT_ENCODE_KEY;
    }

    std::string new_name;
    if (name.length() < KEY_LEN)
    {
        // 用X进行填充
        new_name = std::string(KEY_LEN - name.length(), 'X') + name;
    }
    else
    {
        // 只取前面16个字节
        new_name = name.substr(0, KEY_LEN);
    }

    std::string retval;
    if (!CryptoHelper::des_encrypt(new_name, retval, KEY_LICENSE_KEY))
    {
        KLOG_WARNING("Failed to encrypt string %s by key %s.", new_name.c_str(), KEY_LICENSE_KEY);
        return std::string();
    }
    return retval;
}

std::shared_ptr<ActivationCodeFields> LicenseUtils::parse_activation_code(const std::string &license_code_str,
                                                                          const std::string &key,
                                                                          LicenseErrorCode &error_code)
{
    unsigned char date[DATE_LENGTH];
    std::string activation_code;
    std::string decrypted_activation_code;

    auto result = std::make_shared<ActivationCodeFields>();

    error_code = LicenseUtils::unformat_activation_code(license_code_str, activation_code);
    RETURN_VAL_IF_TRUE(error_code != LicenseErrorCode::LEC_SUCCESS, std::shared_ptr<ActivationCodeFields>());

    if (!CryptoHelper::des_decrypt(activation_code, decrypted_activation_code, key))
    {
        error_code = LicenseErrorCode::LEC_LICENSE_DECODE_FAILED;
        return std::shared_ptr<ActivationCodeFields>();
    }

    result->machine_code = LicenseUtils::format_machine_code(decrypted_activation_code);

    memcpy(date, decrypted_activation_code.data() + MACHINE_CODE_LENGTH, DATE_LENGTH);
    if (!TimeUtils::arr2gdate(date, &result->expired_date))
    {
        error_code = LicenseErrorCode::LEC_LICENSE_DATE_INVALID;
        return std::shared_ptr<ActivationCodeFields>();
    }
    result->customer_info = std::string(decrypted_activation_code.data() + MACHINE_CODE_LENGTH + DATE_LENGTH, CUSTOMER_CODE_LENGTH);
    return result;
}

}  // namespace Kiran
