/**
 * @file          /kylin-license/lib/ukey/ukey-database.cpp
 * @brief         
 * @author        tangjie02 <tangjie02@kylinos.com.cn>
 * @copyright (c) 2020 KylinSec. All rights reserved. 
 */

#include "lib/ukey/ukey-database.h"
#include "lib/base/base.h"

namespace Kiran
{
#define DB_FILE_PERMISSION 0755
#define DB_FILE_PATH KL_INSTALL_DATADIR "/ukeys.db"

#define CREATE_AUTH_TABLE_IF_NO_EXIST                                  \
    "CREATE TABLE IF NOT EXISTS license_authorized_info(\n"            \
    "id                   INTEGER        PRIMARY KEY AUTOINCREMENT,\n" \
    "machine_code         VARCHAR(32)    NOT NULL,\n"                  \
    "license_code         VARCHAR(64)    NOT NULL,\n"                  \
    "expired_date         DATE           NOT NULL,\n"                  \
    "register_time        DATETIME       NOT NULL,\n"                  \
    "hard_id              VARCHAR(32)    NOT NULL,\n"                  \
    "UNIQUE(machine_code, hard_id));\n"

#define CREATE_UKEY_TABLE_IF_NO_EXIST                                      \
    "CREATE TABLE IF NOT EXISTS license_ukey_info(\n"                      \
    "hard_id                           VARCHAR(32)        PRIMARY KEY, \n" \
    "vendor_id                         VARCHAR(32)        NOT NULL,\n"     \
    "customer_info                     VARCHAR(32)        NOT NULL,\n"     \
    "expired_date                      DATE               NOT NULL,\n"     \
    "presence_type                     INT                NOT NULL,\n"     \
    "ukey_license_code               VARCHAR(64),\n"                       \
    "almighty_authorized_points        INT                NOT NULL,\n"     \
    "almighty_remaining_points         INT                NOT NULL,\n"     \
    "desktop_authorized_points         INT                NOT NULL,\n"     \
    "desktop_remaining_points          INT                NOT NULL,\n"     \
    "server_authorized_points          INT                NOT NULL,\n"     \
    "server_remaining_points           INT                NOT NULL,\n"     \
    "industrial_authorized_points      INT                NOT NULL,\n"     \
    "industrial_remaining_points       INT                NOT NULL,\n"     \
    "development_authorized_points     INT                NOT NULL,\n"     \
    "development_remaining_points      INT                NOT NULL);\n"

#define SELECT_ALL_AUTHORIZED_INFO "SELECT id, machine_code, license_code, expired_date, register_time, hard_id from license_authorized_info;"

#define INSERT_AUTH_ITEM                                                 \
    "REPLACE INTO license_authorized_info("                              \
    "machine_code, license_code, expired_date, register_time, hard_id) " \
    "VALUES('{0}', '{1}', DATE('{2}'), DATETIME('{3}'), '{4}');"

#define CLEAR_AUTH_ITEM \
    "DELETE FROM license_authorized_info where hard_id='{0}';"

#define SELECT_ALL_UKEY_INFO                                                             \
    "SELECT hard_id, vendor_id, customer_info, expired_date, presence_type,"             \
    " ukey_license_code, almighty_authorized_points, almighty_remaining_points,"         \
    " desktop_authorized_points, desktop_remaining_points, server_authorized_points,"    \
    "server_remaining_points,industrial_authorized_points, industrial_remaining_points," \
    "development_authorized_points, development_remaining_points from license_ukey_info;"

#define SELECT_UKEY_ITEM_COUNT \
    "SELECT COUNT(*) FROM license_ukey_info WHERE hard_id='%s';"

#define INSERT_UKEY_ITEM                                                                 \
    "REPLACE INTO license_ukey_info("                                                    \
    "hard_id, vendor_id, customer_info, expired_date, presence_type,"                    \
    " ukey_license_code, almighty_authorized_points, almighty_remaining_points,"         \
    " desktop_authorized_points, desktop_remaining_points, server_authorized_points,"    \
    "server_remaining_points,industrial_authorized_points, industrial_remaining_points," \
    "development_authorized_points, development_remaining_points) "                      \
    "VALUES('{0}', '{1}', '{2}', DATE('{3}'), {4}, '{5}', {6}, {7}, {8}, {9}, {10}, {11}, {12}, {13}, {14}, {15});"

#define CLEAR_UKEY_ITEM \
    "DELETE FROM license_ukey_info where presence_type={0};"

UKeyDatabase::UKeyDatabase() : file_path_(DB_FILE_PATH),
                               db_handle_(NULL)
{
}

UKeyDatabase::~UKeyDatabase()
{
}

void UKeyDatabase::init()
{
    KLOG_PROFILE("");

    RETURN_IF_TRUE(this->file_path_.length() == 0);

    if (!Glib::file_test(this->file_path_, Glib::FILE_TEST_EXISTS))
    {
        auto dirname = Glib::path_get_dirname(this->file_path_);
        g_mkdir_with_parents(dirname.c_str(), DB_FILE_PERMISSION);
    }

    // 打开数据库
    auto ret = sqlite3_open(this->file_path_.c_str(), &this->db_handle_);
    if (ret != SQLITE_OK)
    {
        KLOG_WARNING("%s, db_path: %s", sqlite3_errmsg(this->db_handle_), this->file_path_.c_str());
        return;
    }

    if (sqlite3_key(this->db_handle_, DB_KEY, strlen(DB_KEY)) != SQLITE_OK)
    {
        KLOG_WARNING("Failed to decrypt db.");
        return;
    }

    std::string error;

    // 创建授权表
    if (!this->exec_sql(CREATE_AUTH_TABLE_IF_NO_EXIST, error))
    {
        KLOG_WARNING("%s", error.c_str());
    }

    // 创建ukey表
    if (!this->exec_sql(CREATE_UKEY_TABLE_IF_NO_EXIST, error))
    {
        KLOG_WARNING("%s", error.c_str());
    }

    // 清理物理ukey信息
    if (!this->clear_physical_ukeys(error))
    {
        KLOG_WARNING("%s", error.c_str());
    }

    if (!this->load(error))
    {
        KLOG_WARNING("%s", error.c_str());
    }
}

DBUKeyVec UKeyDatabase::get_ukeys(const std::string &vendor_id, int32_t presence_type)
{
    DBUKeyVec result;
    for (auto iter = this->ukey_table_.begin(); iter != this->ukey_table_.end(); ++iter)
    {
        if (iter->second->vendor_id == vendor_id &&
            iter->second->presence_type == presence_type)
        {
            result.push_back(iter->second);
        }
    }
    return result;
}

std::shared_ptr<DBUKeyItem> UKeyDatabase::get_ukey_item(const std::string &hard_id)
{
    auto iter = this->ukey_table_.find(hard_id);
    if (iter != this->ukey_table_.end())
    {
        return iter->second;
    }
    return std::shared_ptr<DBUKeyItem>();
}

bool UKeyDatabase::set_ukey_item(std::shared_ptr<DBUKeyItem> ukey_item)
{
    RETURN_VAL_IF_FALSE(ukey_item, false);

    // 这里做一下简单的校验，剩余点数每次只能减1，避免用户修改内存数据绕过方案
    auto old_ukey_item = this->get_ukey_item(ukey_item->hard_id);
    if (old_ukey_item)
    {
        int32_t old_total_remaining_points = old_ukey_item->almighty_remaining_points +
                                             old_ukey_item->desktop_remaining_points +
                                             old_ukey_item->server_remaining_points +
                                             old_ukey_item->industrial_remaining_points +
                                             old_ukey_item->development_remaining_points;

        int32_t new_total_remaining_points = ukey_item->almighty_remaining_points +
                                             ukey_item->desktop_remaining_points +
                                             ukey_item->server_remaining_points +
                                             ukey_item->industrial_remaining_points +
                                             ukey_item->development_remaining_points;

        // 剩余点数单次增加1个点以上说明有人故意修改数据
        // 这里+1是为了考虑以后存在机器码回收的情况
        if (new_total_remaining_points > old_total_remaining_points + 1)
        {
            KLOG_WARNING("The total remaining points becomes more %d than before", new_total_remaining_points - old_total_remaining_points);
            return false;
        }
    }

    this->ukey_table_[ukey_item->hard_id] = ukey_item;

    auto sql = fmt::format(INSERT_UKEY_ITEM,
                           ukey_item->hard_id,
                           ukey_item->vendor_id,
                           ukey_item->customer_info,
                           ukey_item->expired_date,
                           int(ukey_item->presence_type),
                           ukey_item->ukey_license_code,
                           ukey_item->almighty_authorized_points,
                           ukey_item->almighty_remaining_points,
                           ukey_item->desktop_authorized_points,
                           ukey_item->desktop_remaining_points,
                           ukey_item->server_authorized_points,
                           ukey_item->server_remaining_points,
                           ukey_item->industrial_authorized_points,
                           ukey_item->industrial_remaining_points,
                           ukey_item->development_authorized_points,
                           ukey_item->development_remaining_points);

    std::string error;
    if (!this->exec_sql(sql, error))
    {
        KLOG_WARNING("%s", error.c_str());
        return false;
    }
    return true;
}

DBAuthorizedVec UKeyDatabase::get_authorized_list(const std::string &hard_id)
{
    DBAuthorizedVec result;
    auto iter = this->authorized_table_.find(hard_id);
    if (iter != this->authorized_table_.end())
    {
        result = iter->second;
    }
    return result;
}

static bool authorized_del_cmp(std::shared_ptr<DBAuthorizedItem> authorized_item, std::shared_ptr<DBAuthorizedItem> item)
{
    return authorized_item->hard_id == item->hard_id && authorized_item->machine_code == item->machine_code;
}

bool UKeyDatabase::set_authorized_item(std::shared_ptr<DBAuthorizedItem> authorized_item)
{
    // 删除已有的授权记录
    auto &authorized_list = this->authorized_table_[authorized_item->hard_id];
    auto iter = std::remove_if(authorized_list.begin(), authorized_list.end(), std::bind(authorized_del_cmp, authorized_item, std::placeholders::_1));
    authorized_list.erase(iter, authorized_list.end());
    // 重新添加
    this->authorized_table_[authorized_item->hard_id].push_back(authorized_item);

    auto sql = fmt::format(INSERT_AUTH_ITEM,
                           authorized_item->machine_code,
                           authorized_item->license_code,
                           authorized_item->expired_date,
                           authorized_item->activation_time,
                           authorized_item->hard_id);

    std::string error;
    if (!this->exec_sql(sql, error))
    {
        KLOG_WARNING("%s", error.c_str());
        return false;
    }
    return true;
}

bool UKeyDatabase::clear_physical_ukeys(std::string &error)
{
    std::map<std::string, std::shared_ptr<DBUKeyItem>> ukey_table;

    RETURN_VAL_IF_FALSE(this->exec_sql(SELECT_ALL_UKEY_INFO,
                                       error,
                                       &UKeyDatabase::parse_ukey_item,
                                       &ukey_table),
                        false);

    auto sql = fmt::format(CLEAR_UKEY_ITEM, int(UKeyPresenceType::UKEY_PRESENCE_TYPE_PHYSICAL));
    RETURN_VAL_IF_FALSE(this->exec_sql(sql, error), false);

    // 清理物理ukey已授权信息
    for (auto iter = ukey_table.begin(); iter != ukey_table.end(); ++iter)
    {
        if (iter->second->presence_type == UKeyPresenceType::UKEY_PRESENCE_TYPE_PHYSICAL)
        {
            auto sql = fmt::format(CLEAR_AUTH_ITEM, iter->first);
            RETURN_VAL_IF_FALSE(this->exec_sql(sql, error), false);
        }
    }
    return true;
}

bool UKeyDatabase::load(std::string &error)
{
    this->authorized_table_.clear();
    this->ukey_table_.clear();

    RETURN_VAL_IF_FALSE(this->exec_sql(SELECT_ALL_AUTHORIZED_INFO,
                                       error,
                                       &UKeyDatabase::parse_authorized_item,
                                       &this->authorized_table_),
                        false);

    RETURN_VAL_IF_FALSE(this->exec_sql(SELECT_ALL_UKEY_INFO,
                                       error,
                                       &UKeyDatabase::parse_ukey_item,
                                       &this->ukey_table_),
                        false);
    return true;
}

int UKeyDatabase::parse_authorized_item(void *user_data, int column_num, char **values, char **names)
{
    std::map<std::string, DBAuthorizedVec> *authorized_table = (std::map<std::string, DBAuthorizedVec> *)(user_data);
    RETURN_VAL_IF_TRUE(authorized_table == NULL, -1);
    RETURN_VAL_IF_TRUE(column_num != 6, -1);
    auto authorized_item = std::make_shared<DBAuthorizedItem>();
    // authorized_item->id = strtol(values[0], NULL, 0);
    authorized_item->machine_code = POINTER_TO_STRING(values[1]);
    authorized_item->license_code = POINTER_TO_STRING(values[2]);
    authorized_item->expired_date = POINTER_TO_STRING(values[3]);
    authorized_item->activation_time = POINTER_TO_STRING(values[4]);
    authorized_item->hard_id = POINTER_TO_STRING(values[5]);
    (*authorized_table)[authorized_item->hard_id].push_back(authorized_item);

    return SQLITE_OK;
}

int UKeyDatabase::parse_ukey_item(void *user_data, int column_num, char **values, char **names)
{
    std::map<std::string, std::shared_ptr<DBUKeyItem>> *ukey_table = (std::map<std::string, std::shared_ptr<DBUKeyItem>> *)(user_data);
    RETURN_VAL_IF_TRUE(ukey_table == NULL, -1);
    RETURN_VAL_IF_TRUE(column_num != 16, -1);
    auto ukey_item = std::make_shared<DBUKeyItem>();
    ukey_item->hard_id = POINTER_TO_STRING(values[0]);
    ukey_item->vendor_id = POINTER_TO_STRING(values[1]);
    ukey_item->customer_info = POINTER_TO_STRING(values[2]);
    ukey_item->expired_date = POINTER_TO_STRING(values[3]);
    ukey_item->presence_type = strtol(values[4], NULL, 0);
    ukey_item->ukey_license_code = POINTER_TO_STRING(values[5]);
    ukey_item->almighty_authorized_points = strtol(values[6], NULL, 0);
    ukey_item->almighty_remaining_points = strtol(values[7], NULL, 0);
    ukey_item->desktop_authorized_points = strtol(values[8], NULL, 0);
    ukey_item->desktop_remaining_points = strtol(values[9], NULL, 0);
    ukey_item->server_authorized_points = strtol(values[10], NULL, 0);
    ukey_item->server_remaining_points = strtol(values[11], NULL, 0);
    ukey_item->industrial_authorized_points = strtol(values[12], NULL, 0);
    ukey_item->industrial_remaining_points = strtol(values[13], NULL, 0);
    ukey_item->development_authorized_points = strtol(values[14], NULL, 0);
    ukey_item->development_remaining_points = strtol(values[15], NULL, 0);
    (*ukey_table)[ukey_item->hard_id] = ukey_item;

    return SQLITE_OK;
}

bool UKeyDatabase::exec_sql(std::string sql, std::string &error, sqlite3_callback callback, void *user_data)
{
    RETURN_VAL_IF_TRUE(this->db_handle_ == NULL, true);

    char *sqlite3_error = NULL;
    auto ret = sqlite3_exec(this->db_handle_, sql.c_str(), callback, user_data, &sqlite3_error);
    if (ret != SQLITE_OK)
    {
        error = fmt::format("Failed to exec sql {0}. error: {1}.", sql, sqlite3_error);
        sqlite3_free(sqlite3_error);
        return false;
    }
    KLOG_DEBUG("Success to exec sql: %s.", sql.c_str());
    return true;
}

}  // namespace Kiran
