/**
 * @file          /kylin-license/lib/ukey/ukey-device.cpp
 * @brief         
 * @author        tangjie02 <tangjie02@kylinos.com.cn>
 * @copyright (c) 2020 KylinSec. All rights reserved. 
 */

#include "lib/ukey/ukey-device.h"
#include <glib/gi18n.h>
#include "lib/ukey/ukey-manager.h"

namespace Kiran
{
UKeyDevice::UKeyDevice(const std::string &vendor_id) : vendor_id_(vendor_id)
{
}

void UKeyDevice::print_ukey()
{
    auto &points = this->high_address_area_.points;
    KLOG_INFO("HardID:                      %s", this->hard_id_.c_str());
    // KLOG_DEBUG("    PinCode:                     %s", this->pincode_.c_str());
    KLOG_INFO("CustomerInfo:                %s", this->high_address_area_.customer_info.c_str());
    KLOG_INFO("HardID(2):                   %s", this->high_address_area_.hard_id.c_str());
    KLOG_INFO("ExpredDate:                  %04d-%02d-%02d",
              this->high_address_area_.expire_time.year,
              this->high_address_area_.expire_time.month,
              this->high_address_area_.expire_time.day);
    KLOG_INFO("AlmightyAuthorizedPoints:    %d", points[UKeyPointsIndex::UKEY_POINTS_INDEX_ALMIGHTY].authorized_points);
    KLOG_INFO("AlmightyRemainingPoints:     %d", points[UKeyPointsIndex::UKEY_POINTS_INDEX_ALMIGHTY].remaining_points);
    KLOG_INFO("DesktopAuthorizedPoints:     %d", points[UKeyPointsIndex::UKEY_POINTS_INDEX_DESKTOP].authorized_points);
    KLOG_INFO("DesktopRemainingPoints:      %d", points[UKeyPointsIndex::UKEY_POINTS_INDEX_DESKTOP].remaining_points);
    KLOG_INFO("ServerAuthorizedPoints:      %d", points[UKeyPointsIndex::UKEY_POINTS_INDEX_SERVER].authorized_points);
    KLOG_INFO("ServerRemainingPoints:       %d", points[UKeyPointsIndex::UKEY_POINTS_INDEX_SERVER].remaining_points);
    KLOG_INFO("IndustrialAuthorizedPoints:  %d", points[UKeyPointsIndex::UKEY_POINTS_INDEX_INDUSTRIAL].authorized_points);
    KLOG_INFO("IndustrialRemainingPoints:   %d", points[UKeyPointsIndex::UKEY_POINTS_INDEX_INDUSTRIAL].remaining_points);
    KLOG_INFO("DevelopmentAuthorizedPoints: %d", points[UKeyPointsIndex::UKEY_POINTS_INDEX_DEVELOPMENT].authorized_points);
    KLOG_INFO("DevelopmentRemainingPoints:  %d", points[UKeyPointsIndex::UKEY_POINTS_INDEX_DEVELOPMENT].remaining_points);
}

bool UKeyDevice::is_expired()
{
    auto now_gdate = TimeUtils::get_current_gdate();
    RETURN_VAL_IF_TRUE(g_date_compare(&now_gdate, &this->high_address_area_.expire_time) > 0, true);
    return false;
}

bool UKeyDevice::exist_machine_code(const std::string &machine_code)
{
    auto iter = this->machine_codes_.find(machine_code);
    if (iter == this->machine_codes_.end())
    {
        return false;
    }
    return true;
}

/*
优先查找特定版的剩余点数是否存在，如果都不存在则查找全能版剩余点数
这里的安装形态只能支持最新的枚举类型，老的枚举类型只能使用全能版剩余点数激活了
*/
bool UKeyDevice::has_remaining_points(int32_t install_type)
{
    // 先查找特定版本
    if (install_type >= LicenseInstallType::LICENSE_INSTALL_TYPE_DESKTOP &&
        install_type <= LicenseInstallType::LICENSE_INSTALL_TYPE_DEVELOPMENT &&
        this->high_address_area_.points[install_type - LicenseInstallType::LICENSE_INSTALL_TYPE_DESKTOP + 1].remaining_points > 0)
    {
        return true;
    }

    // 再查找全能版本
    RETURN_VAL_IF_TRUE(this->high_address_area_.points[UKeyPointsIndex::UKEY_POINTS_INDEX_ALMIGHTY].remaining_points > 0, true);

    return false;
}

int32_t UKeyDevice::get_used_points()
{
    int32_t used_points = 0;
    for (int32_t i = 0; i < UKeyPointsIndex::UKEY_POINTS_INDEX_LAST; ++i)
    {
        RETURN_VAL_IF_TRUE(this->high_address_area_.points[i].authorized_points < this->high_address_area_.points[i].remaining_points, -1);
        used_points += (this->high_address_area_.points[i].authorized_points - this->high_address_area_.points[i].remaining_points);
    }
    return used_points;
}

UKeyPointsIndex UKeyDevice::get_remaining_points_index(int32_t install_type)
{
    // 优先使用指定形态的剩余点数，然后使用全能形态的剩余点数
    if (install_type >= LicenseInstallType::LICENSE_INSTALL_TYPE_DESKTOP &&
        install_type <= LicenseInstallType::LICENSE_INSTALL_TYPE_DEVELOPMENT &&
        this->high_address_area_.points[install_type - LicenseInstallType::LICENSE_INSTALL_TYPE_DESKTOP + 1].remaining_points > 0)
    {
        return UKeyPointsIndex(install_type - LicenseInstallType::LICENSE_INSTALL_TYPE_DESKTOP + 1);
    }

    if (this->high_address_area_.points[UKeyPointsIndex::UKEY_POINTS_INDEX_ALMIGHTY].remaining_points > 0)
    {
        return UKeyPointsIndex::UKEY_POINTS_INDEX_ALMIGHTY;
    }
    return UKeyPointsIndex::UKEY_POINTS_INDEX_LAST;
}

UKeyPointsIndex UKeyDevice::get_used_points_index(int32_t install_type)
{
    if (install_type >= LicenseInstallType::LICENSE_INSTALL_TYPE_DESKTOP &&
        install_type <= LicenseInstallType::LICENSE_INSTALL_TYPE_DEVELOPMENT)
    {
        auto index = int(install_type - LicenseInstallType::LICENSE_INSTALL_TYPE_DESKTOP + 1);

        if (this->high_address_area_.points[index].authorized_points > this->high_address_area_.points[index].remaining_points)
        {
            return UKeyPointsIndex(index);
        }
    }

    if (this->high_address_area_.points[UKeyPointsIndex::UKEY_POINTS_INDEX_ALMIGHTY].authorized_points >
        this->high_address_area_.points[UKeyPointsIndex::UKEY_POINTS_INDEX_ALMIGHTY].remaining_points)
    {
        return UKeyPointsIndex::UKEY_POINTS_INDEX_ALMIGHTY;
    }
    return UKeyPointsIndex::UKEY_POINTS_INDEX_LAST;
}

LicenseErrorCode UKeyDevice::add_authorized_info(const std::string &machine_code,
                                                 UKeyPointsIndex &out_points_index)
{
    auto machine_code_str = LicenseUtils::format_machine_code(machine_code);

    RETURN_VAL_IF_TRUE(this->exist_machine_code(machine_code_str), LicenseErrorCode::LEC_UKEY_ALREADY_EXIST_MACHINE_CODE);

    out_points_index = this->get_remaining_points_index(machine_code[0]);
    RETURN_VAL_IF_TRUE(out_points_index == UKeyPointsIndex::UKEY_POINTS_INDEX_LAST, LicenseErrorCode::LEC_UKEY_NO_REMAINING_POINTS_2);

    --this->high_address_area_.points[out_points_index].remaining_points;
    this->machine_codes_.insert(machine_code_str);

    return LicenseErrorCode::LEC_SUCCESS;
}

LicenseErrorCode UKeyDevice::del_authorized_info(const std::string &machine_code, UKeyPointsIndex points_index)
{
    auto machine_code_str = LicenseUtils::format_machine_code(machine_code);

    if (points_index == UKeyPointsIndex::UKEY_POINTS_INDEX_LAST)
    {
        points_index = this->get_used_points_index(machine_code[0]);
    }

    RETURN_VAL_IF_FALSE(this->exist_machine_code(machine_code_str), LicenseErrorCode::LEC_UKEY_NOT_EXIST_MACHINE_CODE);

    RETURN_VAL_IF_TRUE(points_index == UKeyPointsIndex::UKEY_POINTS_INDEX_LAST,
                       LicenseErrorCode::LEC_UKEY_NO_USED_POINTS);

    ++this->high_address_area_.points[points_index].remaining_points;

    this->machine_codes_.erase(machine_code_str);

    return LicenseErrorCode::LEC_SUCCESS;
}

LicenseErrorCode UKeyDevice::sync_authorized_to_db(const std::string &machine_code,
                                                   const std::string &activation_code,
                                                   const GDate &expired_date)
{
    // 同步已授权信息
    auto expired_date_str = TimeUtils::gdate2str(&expired_date);
    auto activation_time = TimeUtils::timestamp2str(time(NULL));
    auto authorized_item = std::make_shared<DBAuthorizedItem>(DBAuthorizedItem{machine_code,
                                                                               activation_code,
                                                                               expired_date_str,
                                                                               activation_time,
                                                                               this->hard_id_});

    auto ukey_database = UKeyManager::get_instance()->get_ukey_database();
    if (!ukey_database->set_authorized_item(authorized_item))
    {
        return LicenseErrorCode::LEC_SYNC_AUTH_TO_DB_FAILED;
    }

    // 同步ukey数据
    expired_date_str = TimeUtils::gdate2str(&this->high_address_area_.expire_time);
    auto &points = this->high_address_area_.points;
    auto ukey_item = std::make_shared<DBUKeyItem>(DBUKeyItem{this->hard_id_,
                                                             this->vendor_id_,
                                                             this->high_address_area_.customer_info,
                                                             expired_date_str,
                                                             UKeyPresenceType::UKEY_PRESENCE_TYPE_VIRTUAL,
                                                             this->high_address_area_.ukey_license_code,
                                                             points[UKeyPointsIndex::UKEY_POINTS_INDEX_ALMIGHTY].authorized_points,
                                                             points[UKeyPointsIndex::UKEY_POINTS_INDEX_ALMIGHTY].remaining_points,
                                                             points[UKeyPointsIndex::UKEY_POINTS_INDEX_DESKTOP].authorized_points,
                                                             points[UKeyPointsIndex::UKEY_POINTS_INDEX_DESKTOP].remaining_points,
                                                             points[UKeyPointsIndex::UKEY_POINTS_INDEX_SERVER].authorized_points,
                                                             points[UKeyPointsIndex::UKEY_POINTS_INDEX_SERVER].remaining_points,
                                                             points[UKeyPointsIndex::UKEY_POINTS_INDEX_INDUSTRIAL].authorized_points,
                                                             points[UKeyPointsIndex::UKEY_POINTS_INDEX_INDUSTRIAL].remaining_points,
                                                             points[UKeyPointsIndex::UKEY_POINTS_INDEX_DEVELOPMENT].authorized_points,
                                                             points[UKeyPointsIndex::UKEY_POINTS_INDEX_DEVELOPMENT].remaining_points});

    if (!ukey_database->set_ukey_item(ukey_item))
    {
        return LicenseErrorCode::LEC_SYNC_UKEY_TO_DB_FAILED;
    }
    return LicenseErrorCode::LEC_SUCCESS;
}

}  // namespace Kiran