
/**
 * @file          /kylin-license/lib/ukey/ukey-group.cpp
 * @brief         
 * @author        tangjie02 <tangjie02@kylinos.com.cn>
 * @copyright (c) 2020 KylinSec. All rights reserved. 
 */

#include "lib/ukey/ukey-group.h"
#include <glib/gi18n.h>
#include <gudev/gudev.h>
#include "lib/ukey/ukey-manager.h"
#include "lib/ukey/ukey-physical-device.h"
#include "lib/ukey/ukey-virtual-device.h"

namespace Kiran
{
#define ROCKEY_SUBSYSTEM "usb"
#define ROCKEY_PROP_PRODUCT_KEY "product"
#define ROCKEY_PROP_PRODUCT_VALUE "ROCKEY3"

#define ROCKEY_NOTIFY_FILE "/etc/kylin-license/rockey_notify"

UKeyGroup::UKeyGroup(const std::string &vendor_id,
                     UKeyPresenceType ukey_presence_type) : vendor_id_(vendor_id),
                                                            ukey_presence_type_(ukey_presence_type),
                                                            udev_client_handler_(0)
{
    const gchar *subsystem[2] = {ROCKEY_SUBSYSTEM, NULL};
    this->udev_client_ = g_udev_client_new(subsystem);
    this->rockey_update_monitor_ = FileUtils::make_monitor_file(ROCKEY_NOTIFY_FILE, sigc::mem_fun(this, &UKeyGroup::on_rockey_file_changed));

    this->udev_client_handler_ = g_signal_connect(this->udev_client_, "uevent", G_CALLBACK(UKeyGroup::on_uevent_changed), this);

    this->load_ukeys();
}

UKeyGroup::~UKeyGroup()
{
    if (this->udev_client_handler_)
    {
        g_signal_handler_disconnect(this->udev_client_, this->udev_client_handler_);
    }

    if (this->udev_client_)
    {
        g_object_unref(this->udev_client_);
    }
}

LicenseErrorCode UKeyGroup::generate_license_code(const std::string &machine_code_str,
                                                  const std::string &key,
                                                  std::shared_ptr<UKeyDevice> &out_ukey_device,
                                                  std::string &out_license_code)
{
    std::string machine_code;

    if (this->ukeys_.size() == 0)
    {
        return LicenseErrorCode::LEC_UKEY_NOT_FOUND;
    }

    RETURN_VAL_IF_ERROR(LicenseUtils::unformat_machine_code(machine_code_str, machine_code));

    auto ukey_device = this->get_ukey_by_machine_code(machine_code_str);

    // 如果该机器没有注册过，或者注册过但是已经过期，则找一个可用的ukey
    if (!ukey_device || ukey_device->is_expired())
    {
        ukey_device = this->find_usable_ukey(machine_code[0]);
        if (!ukey_device)
        {
            return LicenseErrorCode::LEC_UKEY_NOT_USABLE_UKEY_DEVICE;
        }
    }

    RETURN_VAL_IF_ERROR(ukey_device->generate_license_code(machine_code_str, key, out_license_code));
    out_ukey_device = ukey_device;
    return LicenseErrorCode::LEC_SUCCESS;
}

std::shared_ptr<UKeyDevice> UKeyGroup::get_ukey_by_machine_code(const std::string &machine_code)
{
    std::shared_ptr<UKeyDevice> result;
    for (auto iter = this->ukeys_.begin(); iter != this->ukeys_.end(); ++iter)
    {
        auto ukey_device = iter->second;
        if (!ukey_device->exist_machine_code(machine_code))
        {
            continue;
        }

        if (!result)
        {
            result = ukey_device;
            continue;
        }

        // 优先选择物理ukey
        if (result->get_presence_type() == UKeyPresenceType::UKEY_PRESENCE_TYPE_VIRTUAL &&
            ukey_device->get_presence_type() == UKeyPresenceType::UKEY_PRESENCE_TYPE_PHYSICAL)
        {
            result = ukey_device;
            continue;
        }

        // 如果类型相同，优先选择过期时间长的ukey
        if (result->get_presence_type() == ukey_device->get_presence_type() &&
            g_date_compare(&result->get_high_address_area().expire_time, &ukey_device->get_high_address_area().expire_time) < 0)
        {
            result = ukey_device;
            continue;
        }
    }
    return result;
}

UKeyDeviceVec UKeyGroup::get_ukeys()
{
    UKeyDeviceVec result;
    for (auto iter = this->ukeys_.begin(); iter != this->ukeys_.end(); ++iter)
    {
        result.push_back(iter->second);
    }
    return result;
}

void UKeyGroup::load_ukeys()
{
    switch (this->ukey_presence_type_)
    {
    case UKeyPresenceType::UKEY_PRESENCE_TYPE_PHYSICAL:
        this->load_physical_ukeys();
        break;
    case UKeyPresenceType::UKEY_PRESENCE_TYPE_VIRTUAL:
        this->load_virtual_ukeys();
        break;
    default:
        KLOG_WARNING("Unknown ukey presence type.");
        break;
    }
}

void UKeyGroup::load_virtual_ukeys()
{
    KLOG_PROFILE("");

    auto ukey_database = UKeyManager::get_instance()->get_ukey_database();
    auto ukeys = ukey_database->get_ukeys(this->vendor_id_, UKeyPresenceType::UKEY_PRESENCE_TYPE_VIRTUAL);

    for (uint32_t i = 0; i < ukeys.size(); ++i)
    {
        auto &ukey = ukeys[i];
        auto ukey_device = std::make_shared<UKeyVirtualDevice>(ukey);

        if (ukey_device->init())
        {
            this->add_ukey(ukey_device);
        }
    }
}

void UKeyGroup::load_physical_ukeys()
{
    KLOG_PROFILE("");

#if defined(__x86_64__) && defined(ENABLE_ROCKEY)

    unsigned long retcode = 0;
    int32_t ukey_num = 0;

    this->clear_physical_ukeys();

    // 查找设备
    retcode = RY3_Find((char *)this->vendor_id_.c_str(), &ukey_num);
    RETURN_IF_TRUE(retcode != RY3_SUCCESS);

    for (int32_t i = 1; i <= ukey_num; i++)
    {
        RY_HANDLE handle;
        // 打开设备
        retcode = RY3_Open(&handle, i);
        if (retcode != RY3_SUCCESS)
        {
            KLOG_WARNING("Failed to open %d-th USBKey device.", i);
            continue;
        }

        auto ukey_device = std::make_shared<UKeyPhysicalDevice>(this->vendor_id_, handle);
        if (ukey_device->init())
        {
            this->add_ukey(ukey_device);
        }
    }
#endif
}

// 更新UKEY数据
void UKeyGroup::update_physical_ukeys()
{
    for (auto iter = this->ukeys_.begin(); iter != this->ukeys_.end(); ++iter)
    {
        auto ukey_device = iter->second;
        std::string error;

        if (ukey_device->get_presence_type() == UKeyPresenceType::UKEY_PRESENCE_TYPE_PHYSICAL &&
            !ukey_device->update(error))
        {
            KLOG_WARNING("%s", error.c_str());
        }
    }
}

void UKeyGroup::clear_physical_ukeys()
{
    auto iter = this->ukeys_.begin();
    while (iter != this->ukeys_.end())
    {
        auto &ukey = iter->second;
        if (ukey->get_presence_type() == UKeyPresenceType::UKEY_PRESENCE_TYPE_PHYSICAL)
        {
            this->ukeys_.erase(iter++);
        }
        else
        {
            ++iter;
        }
    }
}

void UKeyGroup::add_ukey(std::shared_ptr<UKeyDevice> ukey_device)
{
    RETURN_IF_FALSE(ukey_device);

    if (this->ukeys_.find(ukey_device->get_hard_id()) == this->ukeys_.end())
    {
        this->ukeys_[ukey_device->get_hard_id()] = ukey_device;
    }
    else
    {
        KLOG_WARNING("The hard id '%s' is already exist", ukey_device->get_hard_id().c_str());
    }
}

std::shared_ptr<UKeyDevice> UKeyGroup::find_usable_ukey(int32_t install_type)
{
    // 优先使用物理ukey设备,然后再使用虚拟ukey设备
    for (auto iter = this->ukeys_.begin(); iter != this->ukeys_.end(); ++iter)
    {
        auto ukey_device = iter->second;
        if (ukey_device->get_presence_type() == UKeyPresenceType::UKEY_PRESENCE_TYPE_PHYSICAL &&
            ukey_device->has_remaining_points(install_type) &&
            !ukey_device->is_expired())
        {
            return ukey_device;
        }
    }

    for (auto iter = this->ukeys_.begin(); iter != this->ukeys_.end(); ++iter)
    {
        auto ukey_device = iter->second;
        if (ukey_device->get_presence_type() == UKeyPresenceType::UKEY_PRESENCE_TYPE_VIRTUAL &&
            ukey_device->has_remaining_points(install_type) &&
            !ukey_device->is_expired())
        {
            return ukey_device;
        }
    }
    return std::shared_ptr<UKeyDevice>();
}

void UKeyGroup::on_uevent_changed(void *client, gchar *action, void *device, gpointer user_data)
{
    const gchar *name = NULL;
    const gchar *devtype = NULL;
    UKeyGroup *self = (UKeyGroup *)user_data;

    g_return_if_fail(device != NULL);

    name = g_udev_device_get_name((GUdevDevice *)device);
    devtype = g_udev_device_get_devtype((GUdevDevice *)device);

    KLOG_DEBUG("action: %s name: %s type: %s number: %s sysfs_path: %s.",
               action,
               name,
               devtype,
               g_udev_device_get_number((GUdevDevice *)device),
               g_udev_device_get_sysfs_path((GUdevDevice *)device));

    if (g_strcmp0(action, "add") != 0 && g_strcmp0(action, "remove") != 0)
    {
        return;
    }

    // const gchar* const * props = g_udev_device_get_property_keys(device);
    // for (int i = 0; props[i] != NULL; ++i) {
    //     const gchar *value = g_udev_device_get_property(device, props[i]);
    //     KLOG_DEBUG("property key: %s value: %s.", props[i], value);
    // }

    self->load_physical_ukeys();
    // self->ukey_changed_.emit();
}

void UKeyGroup::on_rockey_file_changed(const Glib::RefPtr<Gio::File> &file,
                                       const Glib::RefPtr<Gio::File> &other_file,
                                       Gio::FileMonitorEvent event_type)
{
    KLOG_DEBUG("event_type: %d.", event_type);

    switch (event_type)
    {
    case Gio::FILE_MONITOR_EVENT_CHANGED:
    case Gio::FILE_MONITOR_EVENT_CREATED:
        this->update_physical_ukeys();
        break;
    default:
        break;
    }
}

}  // namespace Kiran
