/**
 * @file          /kylin-license/lib/ukey/ukey-virtual-device.cpp
 * @brief         
 * @author        tangjie02 <tangjie02@kylinos.com.cn>
 * @copyright (c) 2020 KylinSec. All rights reserved. 
 */

#include "lib/ukey/ukey-virtual-device.h"
#include <dbus-c++/glib-integration.h>
#include <glib/gi18n.h>
#include "lib/base/base.h"
#include "lib/ukey/ukey-manager.h"

namespace Kiran
{
// 虚拟ukey设备每次授权的时间
#define UKEY_VIRTUAL_MAX_AUTHORIZED_DAYS 7

UKeyVirtualDevice::UKeyVirtualDevice(std::shared_ptr<DBUKeyItem> ukey) : UKeyDevice(ukey->vendor_id)
{
    RETURN_IF_FALSE(ukey);
    this->hard_id_ = ukey->hard_id;
    this->pincode_ = this->hard_id_ + PIN_SUFFIX;
    this->high_address_area_.points[UKeyPointsIndex::UKEY_POINTS_INDEX_ALMIGHTY].authorized_points = ukey->almighty_authorized_points;
    this->high_address_area_.points[UKeyPointsIndex::UKEY_POINTS_INDEX_ALMIGHTY].remaining_points = ukey->almighty_remaining_points;
    this->high_address_area_.points[UKeyPointsIndex::UKEY_POINTS_INDEX_DESKTOP].authorized_points = ukey->desktop_authorized_points;
    this->high_address_area_.points[UKeyPointsIndex::UKEY_POINTS_INDEX_DESKTOP].remaining_points = ukey->desktop_remaining_points;
    this->high_address_area_.points[UKeyPointsIndex::UKEY_POINTS_INDEX_SERVER].authorized_points = ukey->server_authorized_points;
    this->high_address_area_.points[UKeyPointsIndex::UKEY_POINTS_INDEX_SERVER].remaining_points = ukey->server_remaining_points;
    this->high_address_area_.points[UKeyPointsIndex::UKEY_POINTS_INDEX_INDUSTRIAL].authorized_points = ukey->industrial_authorized_points;
    this->high_address_area_.points[UKeyPointsIndex::UKEY_POINTS_INDEX_INDUSTRIAL].remaining_points = ukey->industrial_remaining_points;
    this->high_address_area_.points[UKeyPointsIndex::UKEY_POINTS_INDEX_DEVELOPMENT].authorized_points = ukey->development_authorized_points;
    this->high_address_area_.points[UKeyPointsIndex::UKEY_POINTS_INDEX_DEVELOPMENT].remaining_points = ukey->development_remaining_points;
    this->high_address_area_.customer_info = ukey->customer_info;
    this->high_address_area_.ukey_license_code = ukey->ukey_license_code;
    g_date_set_parse(&this->high_address_area_.expire_time, ukey->expired_date.c_str());
    this->high_address_area_.hard_id = ukey->hard_id;

    // 这里是为了防止虚拟ukey拷贝，虚拟ukey中维护了一个激活码，只有激活码跟当前机器的机器码匹配才可以使用该虚拟ukey
    try
    {
        auto dbus_connection = ::DBus::Connection::SystemBus();
        this->license_manager_proxy_ = DBus::LicenseManagerProxyWrapper::create(dbus_connection, LICENSE_MANAGER_OBJECT_PATH, LICENSE_MANAGER_DBUS_NAME);
        auto object_path = this->license_manager_proxy_->GetLicenseObject(LICENSE_OBJECT_VUKEY_NAME);
        this->license_object_proxy_ = DBus::LicenseObjectProxyWrapper::create(dbus_connection, object_path, LICENSE_MANAGER_DBUS_NAME);
        this->license_object_proxy_->ActivateByActivationCode(this->high_address_area_.ukey_license_code);
    }
    catch (const std::exception &e)
    {
        KLOG_WARNING("%s", e.what());
    }
}

bool UKeyVirtualDevice::init()
{
    RETURN_VAL_IF_FALSE(this->hard_id_.length() != 0, false);

    // 计算过期也需要加载进来，便于给到用户正确的错误提示
    // auto now_gdate = TimeUtils::get_current_gdate();
    // RETURN_VAL_AND_ERR_IF_TRUE(g_date_compare(&now_gdate, &this->high_address_area_.expire_time) > 0,
    //                            false,
    //                            error,
    //                            _("The USBKey device is expired, hard id: {0}"),
    //                            this->hard_id_);

    auto ukey_database = UKeyManager::get_instance()->get_ukey_database();
    auto authorized_info = ukey_database->get_authorized_list(this->hard_id_);
    for (uint32_t i = 0; i < authorized_info.size(); ++i)
    {
        auto &authorized_item = authorized_info[i];
        this->machine_codes_.insert(authorized_item->machine_code);
    }

    // 校验授权码是否合法，只有授权码和当前机器匹配才能使用该虚拟ukey设备

    try
    {
        auto license_json = this->license_object_proxy_->GetLicense();
        auto license_value = StrUtils::str2json(license_json);

        if (license_value[LICENSE_JK_ACTIVATION_STATUS].asInt() != LicenseActivationStatus::LAS_ACTIVATED ||
            license_value[LICENSE_JK_SERVICE_STATUS].asInt() != LicenseServiceStatus::LICENSE_SERVICE_STATUS_UNEXPIRED)
        {
            KLOG_WARNING("The virtual USBKey is unactivated.");
            return false;
        }
        this->print_ukey();
    }
    catch (const std::exception &e)
    {
        KLOG_WARNING("%s", e.what());
        return false;
    }

    return true;
}

LicenseErrorCode UKeyVirtualDevice::generate_license_code(const std::string &machine_code_str,
                                                          const std::string &key,
                                                          std::string &out_license_code)
{
    std::string machine_code;
    unsigned char tmp[LICENSE_CODE_LENGTH];
    unsigned char ld[DATE_LENGTH];
    std::string encrypted_activation_code;

    RETURN_VAL_IF_ERROR(LicenseUtils::unformat_machine_code(machine_code_str, machine_code));

    if (!g_date_valid(&this->high_address_area_.expire_time))
    {
        return LicenseErrorCode::LEC_UKEY_EXPIRED_DATE_INVALIDE;
    }

    auto expired_date = TimeUtils::get_current_gdate();
    g_date_add_days(&expired_date, UKEY_VIRTUAL_MAX_AUTHORIZED_DAYS);

    if (g_date_compare(&expired_date, &this->high_address_area_.expire_time) > 0)
    {
        expired_date = this->high_address_area_.expire_time;
    }

    RETURN_VAL_IF_ERROR(TimeUtils::gdate2arr(&expired_date, ld));

    memcpy(tmp, machine_code.data(), MACHINE_CODE_LENGTH);
    memcpy(tmp + MACHINE_CODE_LENGTH, ld, DATE_LENGTH);
    memcpy(tmp + MACHINE_CODE_LENGTH + DATE_LENGTH, this->high_address_area_.customer_info.c_str(), CUSTOMER_CODE_LENGTH);

    if (!CryptoHelper::des_encrypt(std::string((char *)tmp, LICENSE_CODE_LENGTH), encrypted_activation_code, key))
    {
        return LicenseErrorCode::LEC_ENCODE_GENERATE_LICENSE_FAILED;
    }

    RETURN_VAL_IF_FALSE(encrypted_activation_code.length() == LICENSE_CODE_LENGTH, LicenseErrorCode::LEC_GENERATE_LICENSE_LENGTH_INALID);

    // 判断之前是否已经注册过
    if (this->exist_machine_code(machine_code_str))
    {
        KLOG_DEBUG("The machine is already registered.");
    }
    else
    {
        UKeyPointsIndex points_index;
        RETURN_VAL_IF_ERROR(this->add_authorized_info(machine_code, points_index));

        auto license_code_str = LicenseUtils::format_activation_code(encrypted_activation_code);
        RETURN_VAL_IF_ERROR(this->sync_authorized_to_db(machine_code_str, license_code_str, expired_date));
    }

    out_license_code = LicenseUtils::format_activation_code(encrypted_activation_code);
    this->print_ukey();
    return LicenseErrorCode::LEC_SUCCESS;
}

void UKeyVirtualDevice::print_ukey()
{
    this->UKeyDevice::print_ukey();

    KLOG_INFO("AuthorizedMachineCodes:");
    for (auto iter = this->machine_codes_.begin(); iter != this->machine_codes_.end(); ++iter)
    {
        KLOG_INFO("    %s", (*iter).c_str());
    }
}

}  // namespace Kiran
