#include "kiran-license-wrapper.h"
#include "license-dbus-helper.h"

#define KIRAN_LICENSE_HELPER_DBUS_NAME "com.kylinsec.Kiran.LicenseHelper"
#define KIRAN_LICENSE_HELPER_DBUS_OBJECT "/com/kylinsec/Kiran/LicenseHelper"

#define KLOG_ERROR(fmt, ...) g_error("%s: " fmt, __func__, ##__VA_ARGS__)
#define KLOG_WARNING(fmt, ...) g_warning("%s: " fmt, __func__, ##__VA_ARGS__)
#define LOG_MESSAGE(fmt, ...) g_message("%s: " fmt, __func__, ##__VA_ARGS__)
#define KLOG_DEBUG(fmt, ...) g_debug("%s: " fmt, __func__, ##__VA_ARGS__)
#define KLOG_INFO(fmt, ...) g_info("%s: " fmt, __func__, ##__VA_ARGS__)

enum
{
    SIGNAL_STATUS_READY = 0,
    SIGNAL_STATUS_CHANGED = 1,
    SIGNAL_MAX
};

struct _KiranLicenseWrapper
{
    GDBusConnection *system_bus;
    KiranLicenseHelper *proxy;

    KiranLicenseStatus status;
    gboolean is_ready;
    guint dbus_watch;
};

G_DEFINE_TYPE(KiranLicenseWrapper, kiran_license_wrapper, G_TYPE_OBJECT)

static guint64 signals[SIGNAL_MAX] = {0};

/**
 * @brief 解析从DBus后台返回的GVariant格式的授权信息
 * @param license_info  待解析的授权信息
 * @param status            解析出来的授权状态，可以为NULL
 * @param license_code      解析出来的授权码，可以为NULL. 如果提供了，使用后要调用g_free进行释放
 * @param machine_code      解析出来的机器码，可以为NULL. 如果提供了，使用后要调用g_free进行释放
 * @param register_type     解析出来的注册类型，可以为NULL
 * @param register_time     解析出来的注册时间，可以为NULL
 * @param expire_time       解析出来的注册过期时间，可以为NULL
 * @return 解析成功返回TRUE，失败返回FALSE
 */
static gboolean license_info_parse(GVariant *license_info,
                                   KiranLicenseStatus *status,
                                   char **license_code,
                                   char **machine_code,
                                   guint16 *register_type,
                                   guint32 *register_time,
                                   guint32 *expire_time)
{
    KiranLicenseStatus status_ = 0;
    u_int16_t register_type_;
    g_autofree char *license_code_ = NULL, *machine_code_ = NULL;
    u_int32_t register_time_, expire_time_;
    const char *type_string;

    g_return_val_if_fail(license_info != NULL, FALSE);

    type_string = g_variant_get_type_string(license_info);
    g_return_val_if_fail(!strcmp(type_string, "(qssquu)"), FALSE);

    g_variant_get(license_info,
                  "(qssquu)",
                  &status_,
                  &license_code_,
                  &machine_code_,
                  &register_type_,
                  &register_time_,
                  &expire_time_);

#define SET_VAR(x)         \
    do                     \
    {                      \
        if (x)             \
            *(x) = (x##_); \
    } while (0)

#define SET_STRING_VAR(x)            \
    do                               \
    {                                \
        if (x)                       \
            *(x) = g_strdup((x##_)); \
    } while (0)

    SET_VAR(status);
    SET_STRING_VAR(license_code);
    SET_STRING_VAR(machine_code);
    SET_VAR(register_type);
    SET_VAR(register_time);
    SET_VAR(expire_time);

#undef SET_VAR
#undef SET_STRING_VAR

    return TRUE;
}

/**
 * @brief 从给定的license信息中获取授权状态
 * @param license_info 待解析的授权信息
 * @param status       返回解析出来的授权状态，不能为NULL
 * @return 是否成功，成功返回TRUE，是否返回FALSE
 */
static gboolean license_info_to_status(GVariant *license_info, KiranLicenseStatus *status)
{
    guint32 expire_time;
    gboolean expired;
    KiranLicenseStatus status_;

    g_return_val_if_fail(license_info != NULL, FALSE);
    g_return_val_if_fail(status != NULL, FALSE);

    if (!license_info_parse(license_info, &status_, NULL, NULL, NULL, NULL, &expire_time))
    {
        return FALSE;
    }

    expired = (expire_time < time(NULL));
    if (expired)
    {
        switch (status_)
        {
        case LICENSE_STATUS_UNREGISTERED:
            status_ = LICENSE_STATUS_UNREGISTERED_EXPIRED;
            break;
        case LICENSE_STATUS_REGISTERED:
            status_ = LICENSE_STATUS_REGISTERED_EXPIRED;
            break;
        default:
            /* 不需要处理 */
            break;
        }
    }

    *status = status_;
    return TRUE;
}

/**
 * @brief 回调函数: 从后台调用获取授权信息接口返回时调用
 * @see   GAsyncReadyCallback
 * @param proxy
 * @param res
 * @param userdata
 * @return (void)
 */
static void on_license_info_ready(KiranLicenseHelper *proxy,
                                  GAsyncResult *res,
                                  gpointer userdata)
{
    KiranLicenseStatus status;
    GError *error = NULL;
    g_autoptr(GVariant) license_info = NULL;
    KiranLicenseWrapper *self = KIRAN_LICENSE_WRAPPER(userdata);

    g_return_if_fail(proxy != NULL);
    KLOG_DEBUG("Got reply from DBus");
    if (!kiran_license_helper_call_get_license_finish(proxy,
                                                      &license_info,
                                                      res,
                                                      &error))
    {
        /* 通过DBus获取授权信息失败，保持上次请求时的授权状态不变 */
        KLOG_WARNING("Failed to get license status from DBus: %s, use old status %d",
                     error->message, self->status);
        g_error_free(error);
        self->proxy = NULL;

        return;
    }
    else
    {
        if (!license_info_to_status(license_info, &status))
        {
            /* 解析DBus返回的授权信息失败，保持上次请求时的授权状态不变 */
            KLOG_WARNING("Failed to parse license info, use old status %d", self->status);
            return;
        }
    }

    KLOG_DEBUG("new status %d, old %d", status, self->status);
    if (self->status == status)
    {
        KLOG_DEBUG("License status not changed");
        return;
    }

    self->status = status;
    if (!self->is_ready)
    {
        KLOG_DEBUG("emit status-ready signal");
        self->is_ready = TRUE;
        g_signal_emit_by_name(self, "status-ready");
    }
    else
    {
        KLOG_DEBUG("emit status-changed signal");
        g_signal_emit_by_name(self, "status-changed");
    }

    return;
}

/**
 * @brief 通过异步调用DBus接口读取授权状态
 * @param self      License封装对象
 * @return (void)
 */
static void kiran_license_wrapper_read_status(KiranLicenseWrapper *self)
{
    g_return_if_fail(self->proxy != NULL);

    kiran_license_helper_call_get_license(self->proxy,
                                          NULL,
                                          (GAsyncReadyCallback)on_license_info_ready,
                                          self);
}

/**
 * @brief 通过同步调用DBus接口读取授权状态
 * @param self      License封装对象
 * @param status    用于返回读取到的授权状态，不能为NULL
 * @param error     用于返回错误信息，可以为NULL
 * @return 读取成功返回TRUE，失败返回FALSE，并填充 error 信息
 */
gboolean kiran_license_wrapper_read_status_sync(KiranLicenseWrapper *self, KiranLicenseStatus *status, GError **error)
{
    GVariant *license_info;
    GError *error_tmp = NULL;
    KiranLicenseStatus status_;

    if (self->proxy == NULL)
    {
        KLOG_WARNING("Failed to connect to DBus: service not launched");
        error_tmp = g_error_new(G_DBUS_ERROR, G_DBUS_ERROR_NAME_HAS_NO_OWNER, "DBus Service not launched");
        g_propagate_error(error, error_tmp);
        return FALSE;
    }

    if (!kiran_license_helper_call_get_license_sync(self->proxy,
                                                    &license_info,
                                                    NULL,
                                                    &error_tmp))
    {
        KLOG_WARNING("Failed to read license info from DBus: %s", error_tmp->message);
        g_propagate_error(error, error_tmp);
        return FALSE;
    }

    if (!license_info_to_status(license_info, &status_))
    {
        KLOG_WARNING("Failed to parse license info");
        return FALSE;
    }

    *status = status_;
    return TRUE;
}

/**
 * @brief 回调函数: DBus后台发出LicenseChanged信号时调用
 * @param helper            发出信号的DBusProxy
 * @param status_changed    授权状态是否发生变化，该状态只包括了是否授权，不包含过期信息
 * @param user_data         连接信号时提供的用户数据
 * @return (void)
 */
static void on_license_info_changed(KiranLicenseHelper *helper,
                                    gboolean status_changed,
                                    gpointer user_data)
{
    KiranLicenseWrapper *self = NULL;

    KLOG_DEBUG("license file changed");
    self = KIRAN_LICENSE_WRAPPER(user_data);
    /* 重新通过DBus读取授权状态，以确定授权状态是否发生变化 */
    kiran_license_wrapper_read_status(self);
}

/**
 * @brief  连接到授权的DBus后台服务
 *         初始化DBus连接，并在连接成功后调用DBus接口获取授权状态
 * @param self 授权封装对象
 * @return 连接成功返回TRUE, 否则返回FALSE
 */
static gboolean kiran_license_wrapper_connect_to_dbus(KiranLicenseWrapper *self)
{
    GError *error = NULL;

    g_return_val_if_fail(self != NULL, FALSE);
    g_return_val_if_fail(self->proxy == NULL, FALSE);

    KLOG_DEBUG("Connecting to DBus Service ...");

    self->proxy = kiran_license_helper_proxy_new_for_bus_sync(G_BUS_TYPE_SYSTEM,
                                                              0,
                                                              KIRAN_LICENSE_HELPER_DBUS_NAME,
                                                              KIRAN_LICENSE_HELPER_DBUS_OBJECT,
                                                              NULL,
                                                              &error);

    if (!self->proxy)
    {
        KLOG_ERROR("Failed to create dbus proxy: %s", error->message);
        g_error_free(error);
        return FALSE;
    }
    else
    {
        g_signal_connect(self->proxy, "license-changed", G_CALLBACK(on_license_info_changed), self);

        /* schedule callback: 从dbus后台读取授权状态 */
        // g_idle_add((GSourceFunc)&kiran_license_wrapper_read_status, self);
    }
    return TRUE;
}

/**
 * @brief 授权DBus服务可用时的回调函数
 * @see   GBusNameAppearedCallback
 * @param connection
 * @param name
 * @param name_owner
 * @param user_data
 * @return (void)
 */
static void on_dbus_ready(GDBusConnection *connection,
                          const gchar *name,
                          const gchar *name_owner,
                          gpointer user_data)
{
    KiranLicenseWrapper *wrapper = KIRAN_LICENSE_WRAPPER(user_data);

    KLOG_DEBUG("DBus is ready now");
    if (wrapper->proxy != NULL)
    {
        KLOG_DEBUG("DBus already connected, ignore ...");
        return;
    }
    kiran_license_wrapper_connect_to_dbus(wrapper);
}

/**
 * @brief 授权DBus服务关闭时的回调函数
 * @see   GBusNameVanishedCallback
 * @param connection
 * @param name
 * @param user_data
 * @return (void)
 */
static void on_dbus_lost(GDBusConnection *connection,
                         const gchar *name,
                         gpointer user_data)
{
    KiranLicenseWrapper *wrapper = KIRAN_LICENSE_WRAPPER(user_data);

    KLOG_DEBUG("DBus is lost");

    if (wrapper->proxy != NULL)
    {
        if (G_IS_OBJECT(wrapper->proxy))
            g_object_unref(wrapper->proxy);
        wrapper->proxy = NULL;
    }
}

static void kiran_license_wrapper_init(KiranLicenseWrapper *self)
{
    g_return_if_fail(self != NULL);

    self->status = LICENSE_STATUS_UNKNOWN;
    self->is_ready = FALSE;
    self->proxy = NULL;

    /* 监控DBus后台服务状态 */
    self->dbus_watch = g_bus_watch_name(G_BUS_TYPE_SYSTEM,
                                        KIRAN_LICENSE_HELPER_DBUS_NAME,
                                        G_BUS_NAME_WATCHER_FLAGS_AUTO_START,
                                        on_dbus_ready,
                                        on_dbus_lost,
                                        self,
                                        NULL);

    /* 尝试连接到DBus服务 */
    kiran_license_wrapper_connect_to_dbus(self);
}

static void kiran_license_wrapper_finalize(GObject *parent)
{
    KiranLicenseWrapper *self = KIRAN_LICENSE_WRAPPER(parent);

    if (self->proxy)
    {
        g_object_unref(self->proxy);
        self->proxy = NULL;
    }

    if (self->system_bus)
    {
        g_object_unref(self->system_bus);
        self->system_bus = NULL;
    }

    g_bus_unwatch_name(self->dbus_watch);
    G_OBJECT_CLASS(kiran_license_wrapper_parent_class)->finalize(parent);
}

static void kiran_license_wrapper_class_init(KiranLicenseWrapperClass *klass)
{
    signals[SIGNAL_STATUS_READY] = g_signal_new("status-ready",
                                                G_TYPE_FROM_CLASS(klass),
                                                G_SIGNAL_RUN_LAST,
                                                0,
                                                NULL,
                                                NULL,
                                                g_cclosure_marshal_VOID__VOID,
                                                G_TYPE_NONE,
                                                0);

    signals[SIGNAL_STATUS_CHANGED] = g_signal_new("status-changed",
                                                  G_TYPE_FROM_CLASS(klass),
                                                  G_SIGNAL_RUN_LAST,
                                                  0,
                                                  NULL,
                                                  NULL,
                                                  g_cclosure_marshal_VOID__VOID,
                                                  G_TYPE_NONE,
                                                  0);

    G_OBJECT_CLASS(klass)->finalize = kiran_license_wrapper_finalize;
}

KiranLicenseStatus kiran_license_wrapper_get_status(KiranLicenseWrapper *self)
{
    if (!self->is_ready)
    {
        KLOG_WARNING("license status not ready");
        kiran_license_wrapper_read_status(self);
        return LICENSE_STATUS_UNKNOWN;
    }
    return self->status;
}

KiranLicenseStatus kiran_license_wrapper_get_status_sync(KiranLicenseWrapper *self, GError **error)
{
    KiranLicenseStatus status;
    if (!self->is_ready)
    {
        KLOG_DEBUG("license status not ready");
        if (kiran_license_wrapper_read_status_sync(self, &status, error))
        {
            KLOG_DEBUG("Got Reply from DBus, status %d", status);
            self->status = status;
        }
        else
        {
            /* 返回之前上次请求时缓存的授权状态 */
            KLOG_WARNING("Error occured while reading from DBus, use old status %d", self->status);
        }

        self->is_ready = TRUE;
    }

    return self->status;
}

KiranLicenseWrapper *kiran_license_wrapper_new(void)
{
    return g_object_new(KIRAN_LICENSE_TYPE_WRAPPER, NULL);
}

gboolean kiran_license_wrapper_is_ready(KiranLicenseWrapper *self)
{
    return self->is_ready;
}
