
#include "src/daemon/dbus/dbus-license-helper.h"

#include <arpa/inet.h>
#include <fcntl.h>
#include <glib.h>
#include <json/json.h>
#include <stdlib.h>
#include <sys/socket.h>
#include <sys/stat.h>
#include <unistd.h>
#include "lib/base/base.h"
#include "src/daemon/daemon.h"
#include "src/daemon/license-configuration.h"
#include "src/daemon/license-message.pb.h"

#define MOTD_PATH "/etc/motd.d/"
#define MOTD_DIR_PERMISSION 0755
#define MOTD_LICENSE_FILE "/etc/motd.d/license"
#define MOTD_LICENSE_FILE_PERMISSION 0666

namespace Kiran
{
namespace Daemon
{
// 该定义用于/etc/profile.d/license.sh中的翻译
#define SYSTEM_NOT_ACTIVATE N_("System not activated")

DBusLicenseHelper::DBusLicenseHelper(DBus::Connection &connection) : DBus::ObjectAdaptor(connection, LICENSE_HELPER_OBJECT_PATH)
{
}

DBusLicenseHelper::~DBusLicenseHelper()
{
}

void DBusLicenseHelper::init()
{
    this->license_object_ = Daemon::get_instance()->find_license_object(LICENSE_OBJECT_OS_NAME);

    this->license_object_->signal_license_info_changed().connect(sigc::mem_fun(this, &DBusLicenseHelper::on_license_info_changed));

    // 目录或文件不存在则创建
    if (!Glib::file_test(MOTD_PATH, Glib::FILE_TEST_EXISTS))
    {
        auto dirname = Glib::path_get_dirname(MOTD_PATH);
        g_mkdir_with_parents(dirname.c_str(), MOTD_DIR_PERMISSION);
    }
}

::DBus::Struct<uint16_t, std::string, std::string, uint16_t, uint32_t, uint32_t> DBusLicenseHelper::GetLicense()
{
    KLOG_PROFILE("");

    auto license_info = this->license_object_->get_license_info();
    ::DBus::Struct<uint16_t, std::string, std::string, uint16_t, uint32_t, uint32_t> retval;

    if (license_info.licensed == LicenseActivationStatus::LAS_ACTIVATED)
    {
        retval._1 = uint16_t(license_info.licensed);
        retval._2 = license_info.activation_code;
        retval._3 = license_info.machine_code;
        retval._4 = uint16_t(license_info.activation_type);
        retval._5 = license_info.activation_type == 0 ? uint32_t(0) : (uint32_t)license_info.activation_time;
        retval._6 = (uint32_t)license_info.expired_time;
    }
    else
    {
        retval._1 = uint16_t(license_info.licensed);
        retval._2 = std::string();
        retval._3 = license_info.machine_code;
        retval._4 = 0;
        retval._5 = 0;
        retval._6 = (uint32_t)license_info.expired_time;
    }
    return retval;
}

std::string DBusLicenseHelper::GetLicenseJson()
{
    KLOG_PROFILE("");

    auto license_info = this->license_object_->get_license_info();

    Json::Value ukey_value;
    Json::StreamWriterBuilder wbuilder;
    wbuilder["indentation"] = "";

    /*license status, license code, machine code, register type, register time and expired time*/

    ukey_value["license_status"] = license_info.licensed;
    ukey_value["machine_code"] = license_info.machine_code;
    ukey_value["expired_time"] = Json::UInt64(license_info.expired_time);
    ukey_value["start_time"] = Json::UInt64(license_info.first_boot_time);
    ukey_value["install_type"] = Daemon::get_instance()->get_configuration()->get_install_type();
    ukey_value["is_pg"] = Daemon::get_instance()->get_configuration()->is_pg_version();

    if (license_info.licensed == LicenseActivationStatus::LAS_ACTIVATED)
    {
        ukey_value["license_code"] = license_info.activation_code;
        ukey_value["register_type"] = license_info.activation_type;
        ukey_value["register_time"] = license_info.activation_type == 0 ? Json::UInt64(0) : Json::UInt64(license_info.activation_time);
        ukey_value["customer_info"] = license_info.customer_code;
    }
    else
    {
        ukey_value["license_code"] = std::string();
        ukey_value["register_type"] = LicenseActivationType::LICENSE_ACTIVATION_TYPE_UNKNOWN;
        ukey_value["register_time"] = 0;
        ukey_value["customer_info"] = std::string();
    }

    return Json::writeString(wbuilder, ukey_value);
}

int32_t DBusLicenseHelper::GetServiceStatus()
{
    KLOG_PROFILE("");

    auto license_info = this->license_object_->get_license_info();
    return license_info.service_status;
}

std::string DBusLicenseHelper::GetMachineCode()
{
    KLOG_PROFILE("");

    return Daemon::get_instance()->get_machine()->get_machine_code();
}

void DBusLicenseHelper::RegisterByLicenseCode(const std::string &activation_code)
{
    KLOG_PROFILE("Activation code: %s", activation_code.c_str());

    auto error_code = this->license_object_->activate_by_activation_code(activation_code);
    if (error_code != LicenseErrorCode::LEC_SUCCESS)
    {
        THROW_DBUSCXX_ERROR(this->dbus_error_message_, error_code);
    }
}

std::string DBusLicenseHelper::RegisterByUsbKey()
{
    KLOG_PROFILE("");

    std::string error;
    std::shared_ptr<UKeyDevice> ukey_device;

    auto error_code = this->license_object_->activate_by_usb_key(CLIENT_RY3_VENDOR_ID, ukey_device);
    if (error_code != LicenseErrorCode::LEC_SUCCESS)
    {
        THROW_DBUSCXX_ERROR(this->dbus_error_message_, error_code);
    }
    else
    {
        Json::Value ukey_value;
        Json::StreamWriterBuilder wbuilder;
        wbuilder["indentation"] = "";

        const auto &high_address_area = ukey_device->get_high_address_area();

        ukey_value["almighty_authorized_points"] = high_address_area.points[UKeyPointsIndex::UKEY_POINTS_INDEX_ALMIGHTY].authorized_points;
        ukey_value["almighty_remaining_points"] = high_address_area.points[UKeyPointsIndex::UKEY_POINTS_INDEX_ALMIGHTY].remaining_points;
        ukey_value["desktop_authorized_points"] = high_address_area.points[UKeyPointsIndex::UKEY_POINTS_INDEX_DESKTOP].authorized_points;
        ukey_value["desktop_remaining_points"] = high_address_area.points[UKeyPointsIndex::UKEY_POINTS_INDEX_DESKTOP].remaining_points;
        ukey_value["server_authorized_points"] = high_address_area.points[UKeyPointsIndex::UKEY_POINTS_INDEX_SERVER].authorized_points;
        ukey_value["server_remaining_points"] = high_address_area.points[UKeyPointsIndex::UKEY_POINTS_INDEX_SERVER].remaining_points;
        ukey_value["industrial_authorized_points"] = high_address_area.points[UKeyPointsIndex::UKEY_POINTS_INDEX_INDUSTRIAL].authorized_points;
        ukey_value["industrial_remaining_points"] = high_address_area.points[UKeyPointsIndex::UKEY_POINTS_INDEX_INDUSTRIAL].remaining_points;
        ukey_value["development_authorized_points"] = high_address_area.points[UKeyPointsIndex::UKEY_POINTS_INDEX_DEVELOPMENT].authorized_points;
        ukey_value["development_remaining_points"] = high_address_area.points[UKeyPointsIndex::UKEY_POINTS_INDEX_DEVELOPMENT].remaining_points;
        return Json::writeString(wbuilder, ukey_value);
    }
}

void DBusLicenseHelper::RegisterOnLine(const std::string &host)
{
    KLOG_PROFILE("Host: %s", host.c_str());

    auto error_code = this->license_object_->activate_online(host);
    if (error_code != LicenseErrorCode::LEC_SUCCESS)
    {
        THROW_DBUSCXX_ERROR(this->dbus_error_message_, error_code);
    }

    if (!this->license_object_->set_auth_server(host))
    {
        THROW_DBUSCXX_ERROR(this->dbus_error_message_, LicenseErrorCode::LEC_COMMON_IO_ERROR);
    }
}

void DBusLicenseHelper::SetAuthServer(const std::string &host)
{
    KLOG_DEBUG("Host: %s", host.c_str());

    std::string error;
    if (!this->license_object_->set_auth_server(host))
    {
        THROW_DBUSCXX_ERROR(this->dbus_error_message_, LicenseErrorCode::LEC_COMMON_IO_ERROR);
    }
}

void DBusLicenseHelper::on_license_info_changed()
{
    this->LicenseChanged(true);
}

}  // namespace Daemon
}  // namespace Kiran
