/**
 * @file          /kylin-license/src/daemon/license-configuration.cpp
 * @brief         
 * @author        tangjie02 <tangjie02@kylinos.com.cn>
 * @copyright (c) 2020 KylinSec. All rights reserved. 
 */

#include "src/daemon/license-configuration.h"

#include <fcntl.h>
#include <sys/types.h>
#include <unistd.h>

namespace Kiran
{
namespace Daemon
{
// 2.2或者以下版本的license文件路径
#define OLD_LICENSE_FILE "/etc/.license"
// 2.3版本license路径
#define NEW_LICENSE_FILE KL_INSTALL_DATADIR "/license"

#define LICENSE_FILE_PERMISSION 0644
#define MAC_ADDR_LIST_DELIM ","

#define LICENSE_OS_GROUP "os"
#define OS_KEY_INSTALL_CLASS "installclass"
#define OS_KEY_NAME "name"

#define INSTALL_TYPE_FILE_PATH "/etc/.installclass"

LicenseConfiguration::LicenseConfiguration() : file_path_(NEW_LICENSE_FILE)
{
    this->license_config_ = std::make_shared<Glib::KeyFile>();
    this->os_config_ = std::make_shared<LicenseOSConfig>();
}

void LicenseConfiguration::init()
{
    KLOG_PROFILE("");

    this->load_license_config();
    this->load_os_config();

    this->license_monitor_ = FileUtils::make_monitor_file(this->file_path_, sigc::mem_fun(this, &LicenseConfiguration::on_license_file_changed));
}

std::shared_ptr<LicenseCommonGroup> LicenseConfiguration::get_license_common_group()
{
    auto license_common_group = std::make_shared<LicenseCommonGroup>();

    license_common_group->first_boot_time = this->get_string(LICENSE_COMMON_KEYFILE_GROUP, LICENSE_COMMON_KEY_FIRST_BOOT_TIME);

    // 加载MAC地址列表
    std::string mac_addrs_str = this->get_string(LICENSE_COMMON_KEYFILE_GROUP, LICENSE_OBJECT_KEY_MAC_ADDR_LIST);
    if (mac_addrs_str.length() > 0)
    {
        auto mac_addrs = g_strsplit(mac_addrs_str.c_str(), MAC_ADDR_LIST_DELIM, -1);
        for (int32_t i = 0; mac_addrs[i] != NULL; ++i)
        {
            license_common_group->mac_addrs.push_back(mac_addrs[i]);
        }
        g_strfreev(mac_addrs);
    }
    return license_common_group;
}

bool LicenseConfiguration::set_license_common_group(std::shared_ptr<LicenseCommonGroup> group)
{
    RETURN_VAL_IF_FALSE(group, false);

    auto mac_addrs = StrUtils::vector_join(MAC_ADDR_LIST_DELIM, group->mac_addrs);
    this->license_config_->set_string(LICENSE_COMMON_KEYFILE_GROUP, LICENSE_COMMON_KEY_FIRST_BOOT_TIME, group->first_boot_time);
    this->license_config_->set_string(LICENSE_COMMON_KEYFILE_GROUP, LICENSE_OBJECT_KEY_MAC_ADDR_LIST, mac_addrs);

    KLOG_DEBUG("Set license common: first boot time: %s, mac addrs: %s.",
               group->first_boot_time.c_str(),
               mac_addrs.c_str());

    RETURN_VAL_IF_FALSE(this->write_and_encode_data(this->file_path_, CryptoHelper::base64_encode, this->license_config_), false);
    this->group_changed_.emit(LICENSE_COMMON_KEYFILE_GROUP);
    return true;
}

LicenseObjectGroupVec LicenseConfiguration::get_license_object_groups()
{
    LicenseObjectGroupVec license_object_groups;

    std::vector<std::string> groups_name = this->license_config_->get_groups();
    for (uint32_t i = 0; i < groups_name.size(); ++i)
    {
        auto &group_name = groups_name[i];
        if (StrUtils::startswith(group_name, LICENSE_OBJECT_KEYFILE_GROUP))
        {
            auto name = group_name.substr(strlen(LICENSE_OBJECT_KEYFILE_GROUP) + 1);
            auto license_object_group = this->get_license_object_group(name);
            CONTINUE_IF_FALSE(license_object_group);
            license_object_groups.push_back(license_object_group);
        }
    }
    return license_object_groups;
}

std::shared_ptr<LicenseObjectGroup> LicenseConfiguration::get_license_object_group(const std::string &name)
{
    auto group_name = LICENSE_OBJECT_GROUP_NAME(name);
    RETURN_VAL_IF_FALSE(this->license_config_->has_group(group_name), std::shared_ptr<LicenseObjectGroup>());

    auto license_object_group = std::make_shared<LicenseObjectGroup>(name);
    license_object_group->activation_code = this->get_string(group_name, LICENSE_OBJECT_KEY_LICENSE_CODE);
    license_object_group->activation_type = LicenseActivationType(this->get_integer(group_name, LICENSE_OBJECT_KEY_ACTIVATION_TYPE));
    license_object_group->activation_time = this->get_string(group_name, LICENSE_OBJECT_KEY_ACTIVATION_TIME);
    return license_object_group;
}

bool LicenseConfiguration::set_license_object_group(std::shared_ptr<LicenseObjectGroup> group)
{
    RETURN_VAL_IF_FALSE(group, false);

    auto group_name = fmt::format("{0} {1}", LICENSE_OBJECT_KEYFILE_GROUP, group->name);

    this->license_config_->set_string(group_name, LICENSE_OBJECT_KEY_LICENSE_CODE, group->activation_code);
    this->license_config_->set_integer(group_name, LICENSE_OBJECT_KEY_ACTIVATION_TYPE, group->activation_type);
    this->license_config_->set_string(group_name, LICENSE_OBJECT_KEY_ACTIVATION_TIME, group->activation_time);

    KLOG_DEBUG("Set license object %s: activation code: %s, activation type: %d, activation time: %s.",
               group->name.c_str(),
               group->activation_code.c_str(),
               group->activation_type,
               group->activation_time.c_str());

    RETURN_VAL_IF_FALSE(this->write_and_encode_data(this->file_path_, CryptoHelper::base64_encode, this->license_config_), false);
    this->group_changed_.emit(group_name);
    return true;
}

Glib::ustring LicenseConfiguration::get_string(const Glib::ustring &group_name, const Glib::ustring &key)
{
    Glib::ustring retval;
    IGNORE_EXCEPTION(retval = this->license_config_->get_string(group_name, key));
    return retval;
};

int LicenseConfiguration::get_integer(const Glib::ustring &group_name, const Glib::ustring &key)
{
    int retval = 0;
    IGNORE_EXCEPTION(retval = this->license_config_->get_integer(group_name, key));
    return retval;
};

bool LicenseConfiguration::set_string(const Glib::ustring &group_name, const Glib::ustring &key, const Glib::ustring &value)
{
    this->license_config_->set_string(group_name, key, value);
    RETURN_VAL_IF_FALSE(this->write_and_encode_data(this->file_path_, CryptoHelper::base64_encode, this->license_config_), false);
    this->group_changed_.emit(group_name);
    return true;
}

bool LicenseConfiguration::set_integer(const Glib::ustring &group_name, const Glib::ustring &key, int value)
{
    this->license_config_->set_integer(group_name, key, value);
    RETURN_VAL_IF_FALSE(this->write_and_encode_data(this->file_path_, CryptoHelper::base64_encode, this->license_config_), false);
    this->group_changed_.emit(group_name);
    return true;
}

bool LicenseConfiguration::is_pg_version()
{
    RETURN_VAL_IF_FALSE(this->os_config_, false);
    auto name_lowcase = StrUtils::to_lowcase(this->os_config_->name);

    if (name_lowcase.find("pg", 0) != std::string::npos)
    {
        return true;
    }
    return false;
}

LicenseInstallType LicenseConfiguration::get_install_type()
{
    RETURN_VAL_IF_FALSE(this->os_config_, LICENSE_INSTALL_TYPE_UNKNOWN);

    std::string install_class_lowcase = StrUtils::to_lowcase(this->os_config_->install_class);

    if (g_strrstr(install_class_lowcase.c_str(), INSTALL_TYPE_KEYWOARD_DESKTOP) != NULL)
    {
        return LICENSE_INSTALL_TYPE_DESKTOP;
    }
    else if (g_strrstr(install_class_lowcase.c_str(), INSTALL_TYPE_KEYWOARD_SERVER) != NULL)
    {
        return LICENSE_INSTALL_TYPE_SERVER;
    }
    else if (g_strrstr(install_class_lowcase.c_str(), INSTALL_TYPE_KEYWOARD_INDUSTRIAL) != NULL)
    {
        return LICENSE_INSTALL_TYPE_INDUSTRIAL;
    }
    else if (g_strrstr(install_class_lowcase.c_str(), INSTALL_TYPE_KEYWOARD_DEVELOPMENT) != NULL)
    {
        return LICENSE_INSTALL_TYPE_DEVELOPMENT;
    }
    else
    {
        return LICENSE_INSTALL_TYPE_UNKNOWN;
    }
}

void LicenseConfiguration::load_license_config()
{
    KLOG_PROFILE("");

    std::string read_filename = this->file_path_;
    bool sync_to_file = false;

    // 为了兼容2.2或者以下版本，如果新的配置文件不存在，说明是第一次加载配置，因此需要读取旧的配置文件
    if (!Glib::file_test(this->file_path_, Glib::FileTest::FILE_TEST_EXISTS))
    {
        read_filename = OLD_LICENSE_FILE;
        sync_to_file = true;
    }

    auto content = FileUtils::read_bdata(read_filename);
    // 不支持长度小于LICENSE_CODE_LENGTH的版本
    if (content.length() < LICENSE_CODE_LENGTH)
    {
        KLOG_WARNING("The length of license file is less than %d.", LICENSE_CODE_LENGTH);
    }
    else if (content.length() == LICENSE_CODE_LENGTH)
    {
        // 兼容旧的版本，2.0版本中该文件只记录了一个激活码数据。
        this->license_config_->set_string(LICENSE_OBJECT_OS_KEYFILE_GROUP, LICENSE_OBJECT_KEY_LICENSE_CODE, LicenseUtils::format_activation_code(content));
    }
    else
    {
        this->license_config_ = this->read_and_decode_data(read_filename, CryptoHelper::base64_decode);
        // 兼容2.1~2.2的版本，这两个版本的操作系统授权和第一次启动时间都放在[License Client]中
        if (this->license_config_->has_group(LICENSE_CLIENT_KEYFILE_GROUP))
        {
            // get_xxx的第二个参数不使用宏定义的原因是因为该字符串常量是老版本使用的，不应该再更改它。
            auto activation_code = this->get_string(LICENSE_CLIENT_KEYFILE_GROUP, "LicenseCode");
            auto activation_type = this->get_integer(LICENSE_CLIENT_KEYFILE_GROUP, "RegisterType");
            auto activation_time = this->get_string(LICENSE_CLIENT_KEYFILE_GROUP, "RegisterTime");
            auto server_addr = this->get_string(LICENSE_CLIENT_KEYFILE_GROUP, "ServerAddr");
            auto start_time = this->get_string(LICENSE_CLIENT_KEYFILE_GROUP, "StartTime");

            this->license_config_->set_string(LICENSE_OBJECT_OS_KEYFILE_GROUP,
                                              LICENSE_OBJECT_KEY_LICENSE_CODE,
                                              activation_code);
            this->license_config_->set_integer(LICENSE_OBJECT_OS_KEYFILE_GROUP,
                                               LICENSE_OBJECT_KEY_ACTIVATION_TYPE,
                                               activation_type);
            this->license_config_->set_string(LICENSE_OBJECT_OS_KEYFILE_GROUP,
                                              LICENSE_OBJECT_KEY_ACTIVATION_TIME,
                                              activation_time);
            this->license_config_->set_string(LICENSE_OBJECT_OS_KEYFILE_GROUP,
                                              LICENSE_OBJECT_KEY_SERVER_ADDR,
                                              server_addr);
            this->license_config_->set_string(LICENSE_COMMON_KEYFILE_GROUP,
                                              LICENSE_COMMON_KEY_FIRST_BOOT_TIME,
                                              start_time);
            this->license_config_->remove_group(LICENSE_CLIENT_KEYFILE_GROUP);
        }
    }

    // 如果第一次启动时间未设置，则进行添加
    if (this->get_string(LICENSE_COMMON_KEYFILE_GROUP, LICENSE_COMMON_KEY_FIRST_BOOT_TIME).empty())
    {
        this->license_config_->set_string(LICENSE_COMMON_KEYFILE_GROUP, LICENSE_COMMON_KEY_FIRST_BOOT_TIME, TimeUtils::timestamp2str(time(NULL)));
        sync_to_file = true;
    }

    // 配置应该要更新，因此同步到配置文件
    if (sync_to_file)
    {
        this->write_and_encode_data(this->file_path_, CryptoHelper::base64_encode, this->license_config_);
    }
}

static std::string cxx_base64_decode(const std::string &source)
{
    gsize size = 0;
    auto dst = g_base64_decode(source.c_str(), &size);
    std::string retval = std::string((char *)dst, size);
    g_free(dst);
    return retval;
}

void LicenseConfiguration::load_os_config()
{
    KLOG_PROFILE("");

    auto keyfile = this->read_and_decode_data(INSTALL_TYPE_FILE_PATH, cxx_base64_decode);
    RETURN_IF_FALSE(keyfile);
    IGNORE_EXCEPTION(this->os_config_->install_class = keyfile->get_string(LICENSE_OS_GROUP, OS_KEY_INSTALL_CLASS));
    IGNORE_EXCEPTION(this->os_config_->name = keyfile->get_string(LICENSE_OS_GROUP, OS_KEY_NAME));
}

std::shared_ptr<Glib::KeyFile> LicenseConfiguration::read_and_decode_data(const std::string &filename, DecodeFunc decode_func)
{
    gsize encrypted_len = 0;

    try
    {
        std::string encrypted_contents;
        if (Glib::file_test(filename, Glib::FILE_TEST_EXISTS))
        {
            auto decrypted_contents = Glib::file_get_contents(filename);
            encrypted_contents = decode_func(decrypted_contents);
        }

        auto lines = StrUtils::split_lines(encrypted_contents);
        for (uint32_t i = 0; i < lines.size(); ++i)
        {
            KLOG_DEBUG("Read License configuration: %s", lines[i].c_str());
        }

        auto keyfile = std::make_shared<Glib::KeyFile>();
        keyfile->load_from_data(encrypted_contents, Glib::KEY_FILE_NONE);
        return keyfile;
    }
    catch (const Glib::Error &e)
    {
        KLOG_WARNING("%s.", e.what().c_str());
    }
    return std::shared_ptr<Glib::KeyFile>();
}

bool LicenseConfiguration::write_and_encode_data(const std::string &filename,
                                                 EncodeFunc encode_func,
                                                 std::shared_ptr<Glib::KeyFile> keyfile)
{
    try
    {
        auto decrypted_contents = keyfile->to_data();

        auto lines = StrUtils::split_lines(decrypted_contents);
        for (uint32_t i = 0; i < lines.size(); ++i)
        {
            KLOG_DEBUG("Write License configuration: %s", lines[i].c_str());
        }

        auto encrypt_data = encode_func(decrypted_contents);
        RETURN_VAL_IF_FALSE(encrypt_data.length() > 0, false);
        FileUtils::write_bdata(filename, encrypt_data, O_WRONLY | O_CREAT | O_TRUNC | O_SYNC, LICENSE_FILE_PERMISSION);
    }
    catch (const Glib::Error &e)
    {
        KLOG_WARNING("%s.", e.what().c_str());
        return false;
    }
    return true;
}

void LicenseConfiguration::on_license_file_changed(const Glib::RefPtr<Gio::File> &file,
                                                   const Glib::RefPtr<Gio::File> &other_file,
                                                   Gio::FileMonitorEvent event_type)
{
    // 这里是避免被认为删除配置文件，如果文件被删除，则使用缓存数据重新写入
    if (event_type == Gio::FILE_MONITOR_EVENT_DELETED)
    {
        this->write_and_encode_data(this->file_path_, CryptoHelper::base64_encode, this->license_config_);
    }
}

}  // namespace Daemon
}  // namespace Kiran
