/**
 * @file          /kylin-license/src/daemon/license-configuration.h
 * @brief         
 * @author        tangjie02 <tangjie02@kylinos.com.cn>
 * @copyright (c) 2020 KylinSec. All rights reserved. 
 */
#pragma once

#include <giomm.h>

#include "lib/base/base.h"

namespace Kiran
{
namespace Daemon
{
#define LICENSE_COMMON_KEYFILE_GROUP "License Common"
#define LICENSE_COMMON_KEY_FIRST_BOOT_TIME "FirstBootTime"

#define LICENSE_CLIENT_KEYFILE_GROUP "License Client"
#define LICENSE_OBJECT_KEYFILE_GROUP "License Object"
#define LICENSE_OBJECT_OS_KEYFILE_GROUP LICENSE_OBJECT_KEYFILE_GROUP " KylinSecOS"
#define LICENSE_OBJECT_GROUP_NAME(name) fmt::format("{0} {1}", LICENSE_OBJECT_KEYFILE_GROUP, name)
#define LICENSE_OBJECT_KEY_LICENSE_CODE "LicenseCode"
#define LICENSE_OBJECT_KEY_ACTIVATION_TYPE "ActivationType"
#define LICENSE_OBJECT_KEY_ACTIVATION_TIME "ActivationTime"
#define LICENSE_OBJECT_KEY_MAC_ADDR_LIST "MacAddressList"
#define LICENSE_OBJECT_KEY_SERVER_ADDR "ServerAddr"

// 许可证通用信息
struct LicenseCommonGroup
{
    // 服务第一次启动时间，可以作为操作系统安装时间
    std::string first_boot_time;
    // MAC地址列表
    std::vector<std::string> mac_addrs;
};

// 许可对象的配置信息
struct LicenseObjectGroup
{
    LicenseObjectGroup(const std::string &n) : name(n),
                                               activation_type(LicenseActivationType::LICENSE_ACTIVATION_TYPE_UNKNOWN){};
    // 许可对象名称
    std::string name;
    // 激活码
    std::string activation_code;
    // 激活方式
    LicenseActivationType activation_type;
    // 激活时间
    std::string activation_time;
    // 激活服务器地址
    std::string auth_server;
};

typedef std::vector<std::shared_ptr<LicenseObjectGroup>> LicenseObjectGroupVec;

struct LicenseOSConfig
{
    std::string install_class;
    std::string name;
};

class LicenseConfiguration
{
public:
    LicenseConfiguration();
    virtual ~LicenseConfiguration(){};

    void init();

    // 获取/设置授权的通用信息
    std::shared_ptr<LicenseCommonGroup> get_license_common_group();
    bool set_license_common_group(std::shared_ptr<LicenseCommonGroup> group);

    // 获取/设置授权对象的信息
    LicenseObjectGroupVec get_license_object_groups();
    std::shared_ptr<LicenseObjectGroup> get_license_object_group(const std::string &name);
    bool set_license_object_group(std::shared_ptr<LicenseObjectGroup> group);

    //
    Glib::ustring get_string(const Glib::ustring &group_name, const Glib::ustring &key);
    int get_integer(const Glib::ustring &group_name, const Glib::ustring &key);
    bool set_string(const Glib::ustring &group_name, const Glib::ustring &key, const Glib::ustring &value);
    bool set_integer(const Glib::ustring &group_name, const Glib::ustring &key, int value);

    // 获取OS相关属性，这里只能读取不能写入
    std::shared_ptr<LicenseOSConfig> get_os_config()
    {
        return this->os_config_;
    };

    // 判断是否为PG版本，先放在这里，后续可能挪动位置
    bool is_pg_version();

    // 获取系统安装类型
    LicenseInstallType get_install_type();

    // Group变化的信号
    sigc::signal<void, const std::string &> &signal_group_changed() { return this->group_changed_; };

private:
    typedef std::function<std::string(const std::string &)> EncodeFunc;
    typedef std::function<std::string(const std::string &)> DecodeFunc; 

    void load_license_config();
    void load_os_config();

    // 从配置中读取数据并解密
    std::shared_ptr<Glib::KeyFile> read_and_decode_data(const std::string &filename, DecodeFunc decode_func);
    // 将数据加密后写入配置
    bool write_and_encode_data(const std::string &filename,
                               EncodeFunc encode_func,
                               std::shared_ptr<Glib::KeyFile> keyfile);

    void on_license_file_changed(const Glib::RefPtr<Gio::File> &file,
                                 const Glib::RefPtr<Gio::File> &other_file,
                                 Gio::FileMonitorEvent event_type);

private:
    Glib::RefPtr<Gio::FileMonitor> license_monitor_;
    sigc::signal<void, Gio::FileMonitorEvent> license_file_changed_;

    // 授权对象的配置信息
    std::shared_ptr<Glib::KeyFile> license_config_;
    std::shared_ptr<LicenseOSConfig> os_config_;
    // 配置文件路径
    std::string file_path_;
    //
    sigc::signal<void, const std::string &> group_changed_;
};

}  // namespace Daemon
}  // namespace Kiran
