#include "src/daemon/license-machine.h"
#include "src/daemon/daemon.h"

#include <fcntl.h>
#include <glib.h>
#include <gudev/gudev.h>
#include <openssl/evp.h>
#include <openssl/opensslv.h>
#include <stdio.h>
#include <unistd.h>

#include "lib/base/base.h"

namespace Kiran
{
namespace Daemon
{
#define INSTALL_TYPE_FILE_PATH "/etc/.installclass"
#define INSTALL_TYPE_STR_MAX_LEN 256
#define NET_SUBSYSTEM "net"

LicenseMachine::LicenseMachine()
{
    const gchar *subsystem[2] = {NET_SUBSYSTEM, NULL};
    this->udev_client_ = g_udev_client_new(subsystem);
}

LicenseMachine::~LicenseMachine()
{
    if (this->udev_client_handler_)
    {
        g_signal_handler_disconnect(this->udev_client_, this->udev_client_handler_);
    }

    if (this->udev_client_)
    {
       g_object_unref(this->udev_client_);
    }
}

void LicenseMachine::init()
{
    this->load_mac_addrs();

    this->udev_client_handler_ = g_signal_connect(this->udev_client_, "uevent", G_CALLBACK(LicenseMachine::on_uevent_changed), this);
}

std::string LicenseMachine::get_machine_code()
{
    char filename[PATH_MAX];
    unsigned char md[EVP_MAX_MD_SIZE];
    unsigned int md_len;
    std::string machine_code;
    int fd;
    int len;

    EVP_MD_CTX *ctx = NULL;
    char buf[BUFSIZ] = {0};

    const char *dmi_nodes[] = {
        "bios_vendor", "chassis_serial", "chassis_vendor", "product_name",
        "product_serial", "product_uuid", "sys_vendor"};

#if OPENSSL_VERSION_NUMBER < 0x10100000L
    ctx = EVP_MD_CTX_create();
#else
    ctx = EVP_MD_CTX_new();
#endif
    EVP_DigestInit(ctx, EVP_md5());

    for (int i = 0; i < int(sizeof(dmi_nodes) / sizeof(dmi_nodes[0])); i++)
    {
        snprintf(filename, PATH_MAX, "/sys/class/dmi/id/%s", dmi_nodes[i]);

        fd = open(filename, O_RDONLY);
        if (fd < 0)
        {
            continue;
        }

        len = read(fd, buf, 255);
        if (len <= 0 || (len == 1 && buf[0] == '\n'))
        {
            close(fd);
            continue;
        }

        EVP_DigestUpdate(ctx, buf, len);

        close(fd);
    }

    auto license_common_group = Daemon::get_instance()->get_configuration()->get_license_common_group();
    if (license_common_group && license_common_group->mac_addrs.size() > 0)
    {
        EVP_DigestUpdate(ctx, license_common_group->mac_addrs[0].c_str(), license_common_group->mac_addrs[0].length());
    }

    EVP_DigestFinal(ctx, md, &md_len);
#if OPENSSL_VERSION_NUMBER < 0x10100000L
    EVP_MD_CTX_destroy(ctx);
#else
    EVP_MD_CTX_free(ctx);
#endif

    auto install_type = Daemon::get_instance()->get_configuration()->get_install_type();
    machine_code.push_back((char)(install_type));
    machine_code.append(std::string((char *)md, MACHINE_CODE_LENGTH - 1));

    return LicenseUtils::format_machine_code(machine_code);
}

void LicenseMachine::load_mac_addrs()
{
    // 如果本机MAC地址列表和配置中的有交集，则使用配置中的MAC地址列表，否则使用本地MAC地址列表
    auto license_common_group = Daemon::get_instance()->get_configuration()->get_license_common_group();
    auto mac_addrs = this->get_addrs();

    KLOG_DEBUG("Cached mac size: %d, cached macs: %s, mac size: %d, macs: %s.",
               license_common_group ? (int)license_common_group->mac_addrs.size() : 0,
               license_common_group ? StrUtils::vector_join(",", license_common_group->mac_addrs).c_str() : "",
               (int)mac_addrs.size(),
               StrUtils::vector_join(",", mac_addrs).c_str());

    auto join_mac_addrs = StrUtils::vector_intersection(license_common_group->mac_addrs, mac_addrs);
    if (join_mac_addrs.size() == 0)
    {
        license_common_group->mac_addrs = mac_addrs;
        Daemon::get_instance()->get_configuration()->set_license_common_group(license_common_group);
    }
}

std::vector<std::string> LicenseMachine::get_addrs()
{
    std::vector<std::string> result;

    auto devices = g_udev_client_query_by_subsystem(this->udev_client_, "net");
    auto len = g_list_length(devices);

    for (int i = 0; i < (int)len; i++)
    {
        auto device = (GUdevDevice *)g_list_nth_data(devices, i);
        auto address = g_udev_device_get_sysfs_attr(device, "address");

        KLOG_DEBUG("Find mac address: %s, name: %s", address, g_udev_device_get_name(device));

        if (address != NULL &&
            strcmp(address, "00:00:00:00:00:00") != 0 &&
            std::find(result.begin(), result.end(), std::string(address)) == result.end())
        {
            result.push_back(address);
        }
    }

    g_list_free_full(devices, g_object_unref);
    return result;
}

bool LicenseMachine::idle_load_mac_addrs()
{
    this->load_mac_addrs();
    return false;
}


void LicenseMachine::on_uevent_changed(GUdevClient *client, gchar *action, GUdevDevice *device, gpointer user_data)
{
    RETURN_IF_TRUE(device == NULL);

    LicenseMachine *self = (LicenseMachine *)(user_data);

    KLOG_DEBUG("Action: %s name: %s type: %s number: %s sysfs_path: %s.",
               action,
               g_udev_device_get_name(device),
               g_udev_device_get_devtype(device),
               g_udev_device_get_number(device),
               g_udev_device_get_sysfs_path(device));

    // 网卡发生了变化，更新MAC地址列表
    auto idle_time = Glib::MainContext::get_default()->signal_idle();
    idle_time.connect(sigc::mem_fun(self, &LicenseMachine::idle_load_mac_addrs));
}

}  // namespace Daemon
}  // namespace Kiran
