/**
 * @file          /kylin-license/src/daemon/license-object.cpp
 * @brief         
 * @author        tangjie02 <tangjie02@kylinos.com.cn>
 * @copyright (c) 2020 KylinSec. All rights reserved. 
 */

#include "src/daemon/license-object.h"

#include <fmt/format.h>
#include <glib/gi18n.h>
#include <sys/stat.h>
#include <sys/sysinfo.h>
#include "lib/base/base.h"
#include "lib/ukey/ukey-manager.h"
#include "src/daemon/daemon.h"
#include "src/daemon/license-configuration.h"
#include "src/daemon/license-machine.h"
#include "src/daemon/license-message.pb.h"

namespace Kiran
{
namespace Daemon
{
#define DESKTOP_TYPE_MEMORY_SIZE_LIMIT (40UL * 1024UL * 1024UL * 1024UL)

LicenseInfo::LicenseInfo() : licensed(LicenseActivationStatus::LAS_UNACTIVATED),
                             service_status(LicenseServiceStatus::LICENSE_SERVICE_STATUS_INVALID),
                             expired_time(0),
                             activation_type(LicenseActivationType::LICENSE_ACTIVATION_TYPE_UNKNOWN),
                             first_boot_time(time(NULL)),
                             activation_time(0)
{
}

LicenseObject::LicenseObject(const std::string &name) : name_(name)
{
    this->configuration_ = Daemon::get_instance()->get_configuration();
    this->socket_client_ = g_socket_client_new();
    this->key_ = LicenseUtils::generate_key(name);
}

LicenseObject::~LicenseObject()
{
    if (this->socket_client_)
    {
        g_object_unref(this->socket_client_);
    }
}

void LicenseObject::init()
{
    g_socket_client_set_timeout(this->socket_client_, LICENSE_SOCKET_TIMEOUT);

    //每10分钟检查一次授权，timing_check_license为虚函数，所以使用lambda进行封装
    auto timeout = Glib::MainContext::get_default()->signal_timeout();
    this->timer_ = timeout.connect_seconds(sigc::mem_fun(this, &LicenseObject::timing_check_license),
                                           TIMEOUT_IN_SECONDS);

    // 程序启动时检测一次
    this->timing_check_license();

    this->configuration_->signal_group_changed().connect(sigc::mem_fun(this, &LicenseObject::on_license_group_changed));
    this->signal_license_info_changed().connect(sigc::mem_fun(this, &LicenseObject::on_license_info_changed));
}

bool LicenseObject::timeout_check_license()
{
    this->timing_check_license();
    return false;
}

bool LicenseObject::timing_check_license()
{
    KLOG_PROFILE("");

    this->update_license_info();

    // 如果即将过期，则添加一个定时器进行检测和更新授权状态
    auto license_info = this->get_license_info();
    auto expired_remaining_time = license_info.expired_time - time(NULL);
    KLOG_DEBUG("Expired remaining: %d.", expired_remaining_time);
    if (expired_remaining_time > 0 && expired_remaining_time < TIMEOUT_IN_SECONDS &&
        !this->expired_timer_)
    {
        KLOG_DEBUG("The license will expired after %d seconds.", expired_remaining_time);
        auto timeout = Glib::MainContext::get_default()->signal_timeout();
        this->expired_timer_ = timeout.connect_seconds(sigc::mem_fun(this, &LicenseObject::timeout_check_license),
                                                       expired_remaining_time + 3);
    }

    return true;
}

LicenseInfo LicenseObject::get_license_info()
{
    KLOG_PROFILE("");

    LicenseErrorCode error_code;
    LicenseInfo result;

    auto license_common_group = this->configuration_->get_license_common_group();
    auto license_object_group = this->configuration_->get_license_object_group(this->name_);

    do
    {
        BREAK_IF_FALSE(license_object_group);
        result.auth_server = license_object_group->auth_server;

        BREAK_IF_FALSE(license_common_group);
        result.first_boot_time = TimeUtils::str2timestamp(license_common_group->first_boot_time.c_str());

        BREAK_IF_TRUE(license_object_group->activation_code.empty());

        auto license_code_fileds = LicenseUtils::parse_activation_code(license_object_group->activation_code, this->key_, error_code);
        if (!license_code_fileds)
        {
            auto error = LICENSE_ERROR2STR(error_code);
            KLOG_WARNING("%s", error.c_str());
            break;
        }

        auto machine_code = this->get_machine_code();
        if (!this->compare_machine_code(machine_code, license_code_fileds->machine_code))
        {
            KLOG_WARNING("Failed to match machine code.");
            break;
        }

        result.licensed = LicenseActivationStatus::LAS_ACTIVATED;
        result.activation_code = license_object_group->activation_code;
        /* 这里使用计算出来的机器码，而不是通过授权码解密出来的机器码，因为解密出来的机器码可能是通过旧的计算方式计算出来的。
        例如以前的服务器版本机器码是02开头，新的计算方式是0D开头，但这两个机器码都是可以在新版本中使用的*/
        result.machine_code = machine_code;
        result.customer_code = license_code_fileds->customer_info;
        // 由于过期日期是按天为单位的，所以时间戳设置为当天的最后一秒
        result.expired_time = TimeUtils::gdate2timestamp(&license_code_fileds->expired_date) + DAY_IN_SECONDS - 1;
        result.activation_type = license_object_group->activation_type;
        result.activation_time = TimeUtils::str2timestamp(license_object_group->activation_time.c_str());
    } while (false);

    // 未注册的系统拷贝当前机器的机器码，已注册的系统使用授权码中的机器码
    if (result.licensed == LicenseActivationStatus::LAS_UNACTIVATED)
    {
        result.expired_time = result.first_boot_time + EXPIRE_IN_SECONDS + DAY_IN_SECONDS - 1;
        result.machine_code = this->get_machine_code();
    }

    auto current_time = time(NULL);
    auto current_gdate = TimeUtils::get_current_gdate();
    auto current_date = TimeUtils::gdate2timestamp(&current_gdate);

    /* 如果当前时间比开始时间还要小，说明用户手动调整了时间，这种情况在试用期是不允许的
    不过为了避免存在一些误差，这里做一个月的宽限期 */
    if (result.licensed == LicenseActivationStatus::LAS_UNACTIVATED &&
        result.first_boot_time > current_time + 30 * 24 * 3600)
    {
        result.service_status = LicenseServiceStatus::LICENSE_SERVICE_STATUS_INVALID;
    }
    // 因为授权码的过期时间只精确到天，所以这里只能对比天数
    else if (current_date <= result.expired_time)
    {
        result.service_status = LicenseServiceStatus::LICENSE_SERVICE_STATUS_UNEXPIRED;
    }
    else
    {
        result.service_status = LicenseServiceStatus::LICENSE_SERVICE_STATUS_EXPIRED;
    }

    return result;
}

std::string LicenseObject::get_machine_code()
{
    return Daemon::get_instance()->get_machine()->get_machine_code();
}

LicenseErrorCode LicenseObject::check_activation_code(const std::string &activation_code)
{
    KLOG_PROFILE("Activation code: %s.", activation_code.c_str());

    std::string license_machine_code;
    LicenseErrorCode error_code;

    auto machine_code = this->get_machine_code();
    auto license_code_fileds = LicenseUtils::parse_activation_code(activation_code, this->key_, error_code);
    RETURN_VAL_IF_FALSE(license_code_fileds, error_code);

    if (!this->compare_machine_code(machine_code, license_code_fileds->machine_code))
    {
        return LicenseErrorCode::LEC_UNMATCH_MACHINE_CODE;
    }

    RETURN_VAL_IF_ERROR(LicenseUtils::unformat_machine_code(license_code_fileds->machine_code, license_machine_code));

    struct sysinfo info;
    // 用户可能在服务器机器上安装桌面版本的系统，这里根据内存大小进行校验，服务器机器的内存一般大于40G
    // 限制桌面版本只能安装在40G内存以下的机器上
    if (!sysinfo(&info))
    {
        if (license_machine_code[0] == LICENSE_INSTALL_TYPE_DESKTOP &&
            info.totalram >= DESKTOP_TYPE_MEMORY_SIZE_LIMIT)
        {
            return LicenseErrorCode::LEC_INSTALL_TYPE_UNMATCH_MACHINE;
        }
    }

    if (this->is_expired(license_code_fileds->expired_date))
    {
        return LicenseErrorCode::LEC_LICENSE_EXPIRED;
    }
    return LicenseErrorCode::LEC_SUCCESS;
}

LicenseErrorCode LicenseObject::activate_by_activation_code(const std::string &activation_code)
{
    return this->activate_with_type(activation_code, LicenseActivationType::LICENSE_ACTIVATION_TYPE_MANNUAL);
}

LicenseErrorCode LicenseObject::activate_online(const std::string &host)
{
    KLOG_PROFILE("host: %s", host.c_str());

    GSocket *socket = NULL;
    std::string activation_code;
    std::string host_back;

    auto connection = g_socket_client_connect_to_host(this->socket_client_, host.c_str(), LICENSE_TCP_PORT, NULL, NULL);
    if (!connection)
    {
        return LicenseErrorCode::LEC_CONNECT_SERVER_FAILED;
    }
    socket = g_socket_connection_get_socket(connection);
    g_socket_set_timeout(socket, LICENSE_SOCKET_TIMEOUT);

    LicenseErrorCode retval = LicenseErrorCode::LEC_SUCCESS;

    do
    {
        retval = this->request_activation(socket);
        BREAK_IF_TRUE(retval != LicenseErrorCode::LEC_SUCCESS);
        retval = this->response_activation(socket, activation_code);
        BREAK_IF_TRUE(retval != LicenseErrorCode::LEC_SUCCESS);
        retval = this->activate_with_type(activation_code, LicenseActivationType::LICENSE_ACTIVATION_TYPE_ONLINE);
        BREAK_IF_TRUE(retval != LicenseErrorCode::LEC_SUCCESS);
    } while (0);

    return retval;
}

LicenseErrorCode LicenseObject::activate_by_usb_key(const std::string &vendor_id, std::shared_ptr<UKeyDevice> &used_ukey_device)
{
    KLOG_PROFILE("");

    std::string activation_code;
    auto machine_code = this->get_machine_code();

    auto ukey_group = UKeyManager::get_instance()->create_ukey_group(vendor_id);
    RETURN_VAL_IF_FALSE(ukey_group, LicenseErrorCode::LEC_UKEY_GROUP_UNINIT);

    RETURN_VAL_IF_ERROR(ukey_group->generate_license_code(machine_code, this->get_key(), used_ukey_device, activation_code));
    RETURN_VAL_IF_ERROR(this->activate_with_type(activation_code, LicenseActivationType::LICENSE_ACTIVATION_TYPE_OFFLINE));
    return LicenseErrorCode::LEC_SUCCESS;
}

LicenseErrorCode LicenseObject::activate_with_type(const std::string &activation_code, LicenseActivationType activation_type)
{
    KLOG_PROFILE("License code: %s register type: %d.", activation_code.c_str(), activation_type);

    RETURN_VAL_IF_ERROR(this->check_activation_code(activation_code));

    auto license_object_group = this->configuration_->get_license_object_group(this->name_);
    RETURN_VAL_IF_FALSE(license_object_group, LicenseErrorCode::LEC_COMMON_NULL_POINTER);

    license_object_group->activation_code = activation_code;
    license_object_group->activation_type = activation_type;
    license_object_group->activation_time = TimeUtils::timestamp2str(time(NULL));

    if (this->configuration_->set_license_object_group(license_object_group))
    {
        // 记录注册成功的日志信息
        KLOG_INFO("Registration successful. Registeration Type: %d, machine code: %s, license code: %s.",
                  activation_type,
                  this->get_machine_code().c_str(),
                  activation_code.c_str());

        this->activation_success_.emit();
    }
    else
    {
        return LicenseErrorCode::LEC_SYNC_TO_FILE_FAILED;
    }
    return LicenseErrorCode::LEC_SUCCESS;
}

// 对比本地机器码和授权码解密后的机器码是否相等
bool LicenseObject::compare_machine_code(const std::string &local_str, const std::string &license_str)
{
    std::string local_machine_code;
    std::string license_machine_code;
    std::string error;

    RETURN_VAL_IF_FALSE(LicenseUtils::unformat_machine_code(local_str, local_machine_code) == LicenseErrorCode::LEC_SUCCESS, false);
    RETURN_VAL_IF_FALSE(LicenseUtils::unformat_machine_code(license_str, license_machine_code) == LicenseErrorCode::LEC_SUCCESS, false);

    // 后7个字节为实际的机器码
    if (memcmp(local_machine_code.data() + 1, license_machine_code.data() + 1, MACHINE_CODE_LENGTH - 1) != 0)
    {
        return false;
    }

    // 第1个字节为安装类型
    // 为了兼容以前的情况，老的安装类型不参与对比
    if (license_machine_code[0] > LICENSE_INSTALL_TYPE_OLD_LAST &&
        license_machine_code[0] < LICENSE_INSTALL_TYPE_LAST &&
        license_machine_code[0] != local_machine_code[0])
    {
        return false;
    }
    // 未知形态不应该比较成功，用户可能删除或修改了配置文件 （不再限定该条件，会导致旧的系统无法适配成功）
    // RETURN_VAL_IF_TRUE(local_machine_code[0] == LICENSE_INSTALL_TYPE_UNKNOWN, false);

    return true;
}

bool LicenseObject::is_expired(const GDate &expired_date)
{
    auto current_date = TimeUtils::get_current_gdate();
    return (g_date_compare(&current_date, &expired_date) > 0);
}

LicenseErrorCode LicenseObject::request_activation(GSocket *socket)
{
    KLOG_PROFILE("");

    KiranLicenseMessage message;
    GError *gerr = NULL;

    auto machine_code = this->get_machine_code();

    message.mutable_head()->set_id(PROTO_ID_REGISTER_REQ);
    message.mutable_head()->set_error_code(PROTO_NO_ERROR);
    message.mutable_body()->mutable_register_req()->set_machine_code(machine_code);
    auto msg_str = message.SerializeAsString();

    if (g_socket_send(socket, msg_str.c_str(), msg_str.size(), NULL, &gerr) < 0)
    {
        g_error_free(gerr);
        return LicenseErrorCode::LEC_SEND_REQUEST_FAILED;
    }
    return LicenseErrorCode::LEC_SUCCESS;
}

LicenseErrorCode LicenseObject::response_activation(GSocket *socket, std::string &activation_code)
{
    KLOG_PROFILE("");

    gssize read_size = 0;
    GError *gerr = NULL;
    KiranLicenseMessage message;
    read_size = g_socket_receive(socket, this->socket_buf_, LICENSE_SOCKET_DATA_MAX_LEN, NULL, &gerr);

    if (read_size <= 0)
    {
        g_error_free(gerr);
        return LicenseErrorCode::LEC_SERVER_RESPONSE_TIMEOUT;
    }

    message.ParseFromArray(this->socket_buf_, read_size);
    if (message.head().error_code() == PROTO_ERROR_FAILED)
    {
        KLOG_WARNING("%s", message.head().error_desc().c_str());
        return LicenseErrorCode::LEC_SERVER_RESPONSE_ERROR;
    }

    activation_code = message.body().register_rsp().license_code();
    KLOG_DEBUG("Receive activation code: %s.", message.body().register_rsp().license_code().c_str());

    return LicenseErrorCode::LEC_SUCCESS;
}

void LicenseObject::update_license_info()
{
    auto license_info = this->get_license_info();

    if (this->cached_license_info_ != license_info)
    {
        this->cached_license_info_ = license_info;
        this->license_info_changed_.emit();
    }
}

void LicenseObject::idle_check_license()
{
    if (!this->idle_timer_)
    {
        auto idle = Glib::MainContext::get_default()->signal_idle();
        this->idle_timer_ = idle.connect(sigc::mem_fun(this, &LicenseObject::timeout_check_license));
    }
}

void LicenseObject::on_license_group_changed(const std::string &group_name)
{
    auto license_object_group_name = fmt::format("{0} {1}", LICENSE_OBJECT_KEYFILE_GROUP, this->get_name());
    RETURN_IF_TRUE(license_object_group_name != group_name);
    this->idle_check_license();
}

void LicenseObject::on_license_info_changed()
{
    // 如果是未注册，或者是已注册但过期，则向授权服务器重新注册
    auto license_info = this->get_license_info();
    do
    {
        auto auth_server = this->get_auth_server();
        if (auth_server.length() == 0)
        {
            break;
        }

        if (license_info.licensed == LicenseActivationStatus::LAS_UNACTIVATED ||
            (license_info.licensed == LicenseActivationStatus::LAS_ACTIVATED &&
             license_info.service_status == LicenseServiceStatus::LICENSE_SERVICE_STATUS_EXPIRED))
        {
            auto error_code = this->activate_online(auth_server);
            if (error_code != LicenseErrorCode::LEC_SUCCESS)
            {
                auto error = LICENSE_ERROR2STR(error_code);
                KLOG_WARNING("%s", error.c_str());
            }
        }
    } while (0);
}

}  // namespace Daemon
}  // namespace Kiran
