/**
 * @file          /kylin-license/src/daemon/license-object.h
 * @brief         
 * @author        tangjie02 <tangjie02@kylinos.com.cn>
 * @copyright (c) 2020 KylinSec. All rights reserved. 
 */
#pragma once

#include "lib/ukey/ukey-device.h"
#include "src/daemon/license-configuration.h"

namespace Kiran
{
namespace Daemon
{
struct LicenseInfo
{
    LicenseInfo();
    // 激活状态
    LicenseActivationStatus licensed;
    // 服务状态
    LicenseServiceStatus service_status;
    // 激活码
    std::string activation_code;
    // 机器码
    std::string machine_code;
    // 客户代码
    std::string customer_code;
    // 过期时间
    time_t expired_time;
    // 激活方式，如果为LICENSE_ACTIVATION_TYPE_UNKNOWN，则激活时间字段不可用
    LicenseActivationType activation_type;
    // 第一次使用系统的时间
    time_t first_boot_time;
    // 激活时间
    time_t activation_time;
    // 授权服务器地址
    std::string auth_server;

    bool operator==(const LicenseInfo &other)
    {
        return (this->licensed == other.licensed &&
                this->service_status == other.service_status &&
                this->activation_code == other.activation_code &&
                this->expired_time == other.expired_time &&
                this->activation_type == other.activation_type &&
                this->first_boot_time == other.first_boot_time &&
                this->activation_time == other.activation_time &&
                this->auth_server == other.auth_server);
    }

    bool operator!=(const LicenseInfo &other)
    {
        return !(this->operator==(other));
    }
};

class LicenseObject
{
public:
    LicenseObject(const std::string &name);
    virtual ~LicenseObject();

    virtual void init();

    bool timeout_check_license();
    bool timing_check_license();

    // 获取授权对象名称
    std::string get_name() { return this->name_; };

    // 获取授权服务器
    std::string get_auth_server()
    {
        return this->configuration_->get_string(LICENSE_OBJECT_GROUP_NAME(this->name_), LICENSE_OBJECT_KEY_SERVER_ADDR);
    }

    // 设置授权服务器
    bool set_auth_server(const std::string &auth_server)
    {
        return this->configuration_->set_string(LICENSE_OBJECT_GROUP_NAME(this->name_), LICENSE_OBJECT_KEY_SERVER_ADDR, auth_server);
    };

    // 获取授权信息
    LicenseInfo get_license_info();

    // 获取机器码
    std::string get_machine_code();

    // 校验激活码是否合法
    LicenseErrorCode check_activation_code(const std::string &activation_code);

    // 通过激活码激活
    LicenseErrorCode activate_by_activation_code(const std::string &activation_code);

    // 在线激活
    LicenseErrorCode activate_online(const std::string &host);

    // 通过UsbKey注册
    LicenseErrorCode activate_by_usb_key(const std::string &vendor_id, std::shared_ptr<UKeyDevice> &used_ukey_device);

    // 设定激活方式激活
    LicenseErrorCode activate_with_type(const std::string &activation_code, LicenseActivationType activation_type);

    // 获取密钥
    std::string get_key() const { return this->key_; };

    sigc::signal<void> &signal_activation_success() { return this->activation_success_; };
    sigc::signal<void> &signal_license_info_changed() { return this->license_info_changed_; };

private:
    bool compare_machine_code(const std::string &local_str, const std::string &license_str);
    bool is_expired(const GDate &expired_date);

    bool is_expired(GDate *expired_date);
    LicenseErrorCode request_activation(GSocket *socket);
    LicenseErrorCode response_activation(GSocket *socket, std::string &activation_code);

    // 更新授权信息
    void update_license_info();

    // 空闲时检查license状态变化情况
    void idle_check_license();

    void on_license_group_changed(const std::string &group_name);
    void on_license_info_changed();

protected:
    // 授权对象的名称
    std::string name_;
    std::shared_ptr<LicenseConfiguration> configuration_;

private:
    // 密钥
    std::string key_;

    sigc::signal<void> activation_success_;
    sigc::signal<void> license_info_changed_;

    GSocketClient *socket_client_;
    // socket通信缓存
    char socket_buf_[LICENSE_SOCKET_DATA_MAX_LEN];

    // 定时器，用于定时检查授权情况
    sigc::connection timer_;
    // 授权已过期时检查授权状态
    sigc::connection expired_timer_;
    // 空闲时检查授权状态
    sigc::connection idle_timer_;

    LicenseInfo cached_license_info_;

    std::vector<std::string> mac_addrs_;
};

typedef std::vector<std::shared_ptr<LicenseObject>> LicenseObjectVec;
}  // namespace Daemon
}  // namespace Kiran
