/**
 * @file          /kylin-license/src/notify/kiran-notify.cpp
 * @brief         
 * @author        tangjie02 <tangjie02@kylinos.com.cn>
 * @copyright (c) 2020 KylinSec. All rights reserved. 
 */

#include "src/notify/kiran-notify.h"

#include <fcntl.h>
#include <glib/gi18n.h>
#include <json/json.h>
#include <sys/stat.h>

#include "lib/base/base.h"
#include "src/notify/license-notify.h"

namespace Kiran
{
KiranNotify::KiranNotify() : check_count_(0)
{
}

KiranNotify::~KiranNotify()
{
    if (this->timer_)
    {
        this->timer_.disconnect();
    }
}

KiranNotify* KiranNotify::instance_ = NULL;

void KiranNotify::global_init()
{
    instance_ = new KiranNotify();
    instance_->init();
}

void KiranNotify::init()
{
    try
    {
        auto dbus_connection = ::DBus::Connection::SystemBus();
        this->proxy_ = DBus::LicenseHelperProxyWrapper::create(dbus_connection, LICENSE_HELPER_OBJECT_PATH, LICENSE_HELPER_DBUS_NAME);
        // TODO:
        // this->proxy_->dbusProxy()->set_default_timeout(LICENSE_DBUS_TIMEOUT_MSEC);
    }
    catch (const Glib::Error& e)
    {
        KLOG_WARNING("%s", e.what().c_str());
    }

    auto timeout = Glib::MainContext::get_default()->signal_timeout();
    this->timer_ = timeout.connect_seconds(sigc::mem_fun(this, &KiranNotify::timing_check_license), TIMEOUT_IN_SECONDS);
    // 启动时检测一次
    this->timing_check_license();
}

bool KiranNotify::timing_check_license()
{
    ++this->check_count_;

    RETURN_VAL_IF_FALSE(this->proxy_, true);

    std::string license_json;
    LicenseServiceStatus service_status;

    try
    {
        license_json = this->proxy_->GetLicenseJson();
        service_status = LicenseServiceStatus(this->proxy_->GetServiceStatus());
    }
    catch (const Glib::Error& e)
    {
        KLOG_WARNING("%s", e.what().c_str());
        return true;
    }

    Json::CharReaderBuilder reader_builder;
    Json::Value license_value;
    std::string error;

    std::unique_ptr<Json::CharReader> reader(reader_builder.newCharReader());

    if (!reader->parse(license_json.c_str(), license_json.c_str() + license_json.length(), &license_value, &error))
    {
        KLOG_WARNING("%s", error.c_str());
        return true;
    }

    auto license_status = license_value["license_status"].asUInt();

    if (license_status == LicenseActivationStatus::LAS_UNACTIVATED)
    {
        // 试用期结束
        if (service_status != LicenseServiceStatus::LICENSE_SERVICE_STATUS_UNEXPIRED)
        {
            // PG版本需要添加重启提示（重启功能已经删除，不再需要提示重启）
            // if (license_value["is_pg"].asBool())
            // {
            //     NOTIFY_WARN(_("The system trial period has expired, Your system will restart in half an hour. We will not response to the loss! Please purchase a license in time\n"));
            // }
            // else
            {
                NOTIFY_WARN(_(
                    "Welcome to use KylinSec OS! Your system is not activated. Please activate your system to enjoy the first-class and professional technical support from Kylinsec team.\n"
                    "Contact us: 400-625-6606\n"));
            }
        }
        // 试用期未结束
        else
        {
            if (this->check_count_ == 1)
            {
                NOTIFY_INFO(_(
                    "Welcome to use KylinSec OS! Your system is not activated. Please activate your system to enjoy the first-class and professional technical support from Kylinsec team.\n"
                    "Contact us: 400-625-6606\n"));
            }
        }
    }
    else if (license_status == LicenseActivationStatus::LAS_ACTIVATED)
    {
        if (this->check_count_ == 1)
        {
            // 已注册，但过期
            if (service_status != LicenseServiceStatus::LICENSE_SERVICE_STATUS_UNEXPIRED)
            {
                NOTIFY_WARN(_(
                    "The warranty period of your license for KylinSec OS has expired. Please purchase the technical service again.\n"
                    "Contact us: 400-625-6606\n"));
            }
            else
            {
                NOTIFY_INFO(_("Welcome to Kylinsec OS.\n"));
            }
        }
    }
    return true;
}

}  // namespace Kiran
