/**
 * @file          /kylin-license/src/register/license-register.cpp
 * @brief         
 * @author        tangjie02 <tangjie02@kylinos.com.cn>
 * @copyright (c) 2020 KylinSec. All rights reserved. 
 */

#include "src/register/license-register.h"

#include <glib/gi18n.h>

namespace Kiran
{
LicenseRegister::LicenseRegister(::DBus::Connection dbus_connection) : option_group_("config", "config options"),
                                                                       dbus_connection_(dbus_connection)
{
}

LicenseRegister::~LicenseRegister()
{
}

bool LicenseRegister::init()
{
    this->option_group_.add_entry(MiscUtils::create_option_entry('r', "register",
                                                                 N_("Activate the machine. If the activation code is specified, then activate machine using activation code,"
                                                                    " and if the server host is specified, then activate machine online. Otherwise, activate using USBKey device.")),
                                  this->options_.registration);
    this->option_group_.add_entry(MiscUtils::create_option_entry('l', "license-code",
                                                                 N_("Specify the activation code, the argument is deprecated, please use -a.")),
                                  this->options_.license_code);
    this->option_group_.add_entry(MiscUtils::create_option_entry('a', "activation-code",
                                                                 N_("Specify the activation code.")),
                                  this->options_.activation_code);
    this->option_group_.add_entry(MiscUtils::create_option_entry('d', "server-host",
                                                                 N_("Specify the remote activation server host.")),
                                  this->options_.server_host);
    this->option_group_.add_entry(MiscUtils::create_option_entry('u', "usbkey-vendor-id",
                                                                 N_("Specify the vendor ID of USBKey.")),
                                  this->options_.vendor_id);
    this->option_group_.add_entry(MiscUtils::create_option_entry('s', "show",
                                                                 N_("Show activation info.")),
                                  this->options_.show_license);
    this->option_group_.add_entry(MiscUtils::create_option_entry('v', "version",
                                                                 N_("Output version information and exit.")),
                                  this->options_.show_version);
    this->option_group_.add_entry(MiscUtils::create_option_entry('g', "get",
                                                                 N_("Get the specified variable."), "registration-status|status|expired-status|machinecode|customercode|authcode|expiredate"),
                                  this->options_.variable_name);
    this->option_group_.add_entry(MiscUtils::create_option_entry('o', "activation-object-name",
                                                                 N_("Name of activation object. If not specified, KylinSecOS is used.")),
                                  this->options_.license_object_name);

    this->option_group_.add_entry(MiscUtils::create_option_entry('j', "json",
                                                                 N_("Use JSON output format.")),
                                  this->options_.show_by_json);

    this->option_group_.set_translation_domain(PROJECT_NAME);
    this->option_context_.set_main_group(this->option_group_);

    try
    {
        this->license_manager_proxy_ = DBus::LicenseManagerProxyWrapper::create(this->dbus_connection_, LICENSE_MANAGER_OBJECT_PATH, LICENSE_MANAGER_DBUS_NAME);
        // TODO:
        // this->license_manager_proxy_->dbusProxy()->set_default_timeout(LICENSE_DBUS_TIMEOUT_MSEC);
    }
    catch (const std::exception& e)
    {
        KLOG_WARNING("%s", e.what());
        fmt::print(stderr, "{0}", e.what());
        return false;
    }

    return true;
}

bool LicenseRegister::run(int& argc, char**& argv)
{
    try
    {
        this->option_context_.parse(argc, argv);
    }
    catch (const Glib::Exception& e)
    {
        KLOG_WARNING("%s", e.what().c_str());
        fmt::print(stderr, _("Failed to parse command arguments\n"));
        return false;
    }

    if (this->options_.show_version)
    {
        fmt::print("Kylin license 2.1\n");
        return true;
    }

    // 兼容旧的版本
    if (this->options_.license_object_name.empty())
    {
        this->options_.license_object_name = LICENSE_OBJECT_OS_NAME;
    }

    // 兼容旧的版本
    if (this->options_.vendor_id.empty())
    {
        this->options_.vendor_id = CLIENT_RY3_VENDOR_ID;
    }

    try
    {
        auto object_path = this->license_manager_proxy_->GetLicenseObject(this->options_.license_object_name);
        this->license_object_proxy_ = DBus::LicenseObjectProxyWrapper::create(this->dbus_connection_, object_path, LICENSE_MANAGER_DBUS_NAME);
        if (!this->options_.variable_name.empty())
        {
            this->process_get_variable(this->options_.variable_name);
            return true;
        }

        if (this->options_.registration)
        {
            // -a和-l选项相同，优先使用-a选项，如果-a未设置但-l设置了才使用-l选项的参数
            if (!this->options_.license_code.empty() && this->options_.activation_code.empty())
            {
                this->options_.activation_code = this->options_.license_code;
            }

            if (!this->process_registration(this->options_.activation_code,
                                            this->options_.server_host,
                                            this->options_.vendor_id))
            {
                return false;
            }
        }

        if (this->options_.show_license)
        {
            this->process_show();
        }
    }
    catch (const std::exception& e)
    {
        fmt::print(stderr, "{0}", e.what());
        return false;
    }

    return true;
}

/**
 * @brief kyreg -g 提供给云桌面用的接口
 * @param variable_name
 */
void LicenseRegister::process_get_variable(const std::string& variable_name)
{
    KLOG_PROFILE("variable name: %s", variable_name.c_str());

    Json::Value license_value;
    std::string error;

    if (!this->get_license(license_value, error))
    {
        fmt::print(stderr, "{0}", error);
        return;
    }

    auto activation_status = license_value["activation_status"].asUInt();
    auto service_status = license_value["service_status"].asUInt();
    auto machine_code = license_value["machine_code"].asString();
    auto customer_info = license_value["customer_info"].asString();
    auto activation_code = license_value["activation_code"].asString();
    auto expired_time = license_value["expired_time"].asUInt64();
    auto expired_gdate = TimeUtils::timestamp2gdate(expired_time);
    auto expired_date_str = TimeUtils::gdate2str(&expired_gdate);

    if (variable_name == "registration-status")
    {
        fmt::print("{0}\n", activation_status == LicenseActivationStatus::LAS_UNACTIVATED ? "unregistered" : "registered");
    }
    else if (variable_name == "status")
    {
        fmt::print("{0}\n", activation_status == LicenseActivationStatus::LAS_UNACTIVATED ? "inactive" : "activate");
    }
    else if (variable_name == "expired-status")
    {
        if (service_status == LicenseServiceStatus::LICENSE_SERVICE_STATUS_UNEXPIRED)
        {
            fmt::print("unexpired\n");
        }
        else
        {
            fmt::print("expired\n");
        }
    }
    else if (variable_name == "machinecode")
    {
        fmt::print("{0}\n", machine_code);
    }
    else if (variable_name == "customercode")
    {
        fmt::print("{0}\n", customer_info);
    }
    else if (variable_name == "authcode")
    {
        fmt::print("{0}\n", activation_code);
    }
    else if (variable_name == "expiredate")
    {
        fmt::print("{0}\n", (g_date_year(&expired_gdate) >= 2250) ? std::string("Forever") : expired_date_str);
    }
}

bool LicenseRegister::process_registration(const std::string& activation_code,
                                           const std::string& auth_server,
                                           const std::string& vendor_id)
{
    KLOG_PROFILE("License code: %s auth server: %s",
                 activation_code.c_str(),
                 auth_server.c_str());

    std::string error;

    RETURN_VAL_IF_FALSE(this->license_object_proxy_, false);

    try
    {
        // 使用激活码激活
        if (activation_code.length() > 0)
        {
            this->license_object_proxy_->ActivateByActivationCode(activation_code);
        }
        // 在线激活
        else if (auth_server.length() > 0)
        {
            this->license_object_proxy_->ActivateOnLine(auth_server);
        }
        // UKEY激活
        else
        {
            auto result = this->license_object_proxy_->ActivateByUsbKey(vendor_id);

            Json::Value ukey_value;

            if (!this->json_str2value(result, ukey_value, error))
            {
                fmt::print(stderr, "{0}", error);
                return false;
            }

            // 兼容旧的逻辑，分两种情况打印
            if (ukey_value["almighty_authorized_points"].asInt() > 0)
            {
                fmt::print(_("Activation points:     {0}\n"), ukey_value["almighty_authorized_points"].asInt());
                fmt::print(_("Remaining points:      {0}\n"), ukey_value["almighty_remaining_points"].asInt());
            }
            else
            {
                fmt::print(_("Desktop activation points:     {0}\n"), ukey_value["desktop_authorized_points"].asInt());
                fmt::print(_("Desktop remaining points:      {0}\n"), ukey_value["desktop_remaining_points"].asInt());
                fmt::print(_("Server activation points:      {0}\n"), ukey_value["server_authorized_points"].asInt());
                fmt::print(_("Server remaining points:       {0}\n"), ukey_value["server_remaining_points"].asInt());
                fmt::print(_("Industrial activation points:  {0}\n"), ukey_value["industrial_authorized_points"].asInt());
                fmt::print(_("Industrial remaining points:   {0}\n"), ukey_value["industrial_remaining_points"].asInt());
                fmt::print(_("Development activation points: {0}\n"), ukey_value["development_authorized_points"].asInt());
                fmt::print(_("Development remaining points:  {0}\n"), ukey_value["development_remaining_points"].asInt());
            }
        }
    }
    catch (const Glib::Error& e)
    {
        fmt::print(stderr, "{0}", e.what().c_str());
        return false;
    }
    return true;
}

void LicenseRegister::process_show()
{
    Json::Value license_value;
    std::string error;

    if (!this->get_license(license_value, error))
    {
        fmt::print(stderr, "{0}", error);
        return;
    }

    // 这里主要是提供给上层应用使用，方便上层应用通过命令行获取激活信息
    if (this->options_.show_by_json)
    {
        auto json_str = StrUtils::json2str(license_value);
        fmt::print("{0}", json_str.c_str());
        return;
    }

    auto activation_status = license_value["activation_status"].asUInt();
    auto service_status = license_value["service_status"].asUInt();
    auto activation_type = license_value["activation_type"].asUInt();
    auto activation_time = license_value["activation_time"].asUInt64();
    auto expired_time = license_value["expired_time"].asUInt64();
    auto first_boot_time = license_value["first_boot_time"].asUInt64();
    auto machine_code = license_value["machine_code"].asString();
    auto activation_code = license_value["activation_code"].asString();
    auto customer_info = license_value["customer_info"].asString();

    auto expired_gdate = TimeUtils::timestamp2gdate(expired_time);
    auto expired_date_str = TimeUtils::gdate2str(&expired_gdate);
    if (activation_status == LicenseActivationStatus::LAS_ACTIVATED)
    {
        auto register_type_str = this->get_activation_type_str(activation_type);

        auto register_date_str = TimeUtils::timestamp2str(activation_time);
        fmt::print(_("Activation status: Activated.\n"));
        fmt::print(_("Expiry date:       {0}.\n"), (g_date_year(&expired_gdate) >= 2250) ? _("Forever") : expired_date_str);
        fmt::print(_("Activation type:   {0}.\n"), register_type_str);
        fmt::print(_("Activation code:   {0}.\n"), activation_code);
        fmt::print(_("Machine code:      {0}.\n"), machine_code);
        fmt::print(_("Activation time:   {0}.\n"), register_date_str);
    }
    else
    {
        if (service_status == LicenseServiceStatus::LICENSE_SERVICE_STATUS_INVALID)
        {
            fmt::print(_("Activation status: The current time is illegal. (less than installation time: {0}).\n"),
                       TimeUtils::timestamp2str(first_boot_time));
        }
        else
        {
            fmt::print(_("Activation status: Not activated.\n"));
            // 只有系统激活才显示试用期
            if (this->options_.license_object_name == LICENSE_OBJECT_OS_NAME)
            {
                fmt::print(_("Trial expiration:  {0}.\n"), expired_date_str);
            }
        }

        fmt::print(_("Machine code:      {0}.\n"), machine_code);
    }

    // 只有系统激活才显示安装时间
    if (this->options_.license_object_name == LICENSE_OBJECT_OS_NAME)
    {
        auto start_date_str = TimeUtils::timestamp2str(first_boot_time);
        fmt::print(_("Installation time: {0}.\n"), start_date_str);
    }
}

bool LicenseRegister::get_license(Json::Value& value, std::string& error)
{
    KLOG_PROFILE("");

    if (!this->license_object_proxy_)
    {
        error = _("The dbus proxy of license object is uninitialized");
        return false;
    }

    try
    {
        auto license_json = this->license_object_proxy_->GetLicense();
        return this->json_str2value(license_json, value, error);
    }
    catch (const Glib::Error& e)
    {
        error = e.what();
        return false;
    }
}

bool LicenseRegister::json_str2value(const std::string& str, Json::Value& value, std::string& error)
{
    Json::CharReaderBuilder reader_builder;
    std::unique_ptr<Json::CharReader> reader(reader_builder.newCharReader());
    return reader->parse(str.c_str(), str.c_str() + str.length(), &value, &error);
}

std::string LicenseRegister::get_activation_type_str(uint32_t activation_type)
{
    switch (activation_type)
    {
    case LicenseActivationType::LICENSE_ACTIVATION_TYPE_MANNUAL:
        return std::string(_("Activation code"));

    case LicenseActivationType::LICENSE_ACTIVATION_TYPE_ONLINE:
        return std::string(_("Network"));

    case LicenseActivationType::LICENSE_ACTIVATION_TYPE_OFFLINE:
        return std::string(_("USBKey"));

    default:
        return std::string(_("Unknown"));
    }
    return std::string(_("Unknown"));
}
}  // namespace Kiran
