#include "src/server/socket-manager.h"
#include <glib/gi18n.h>
#include "lib/ukey/ukey-manager.h"

namespace Kiran
{
#define LICENSE_SERVER_BACKLOG 1024

struct write_req_t
{
    uv_write_t req;
    uv_buf_t buf;
};

SocketManager::SocketManager() : dbus_connection_(::DBus::Connection::SystemBus())
{
}

SocketManager::~SocketManager()
{
}

SocketManager *SocketManager::instance_ = nullptr;
void SocketManager::global_init()
{
    auto dbus_connection = ::DBus::Connection::SystemBus();
    instance_ = new SocketManager();
    instance_->init();
}

void alloc_client_buffer(uv_handle_t *handle, size_t suggested_size,
                         uv_buf_t *buf)
{
    buf->base = (char *)malloc(suggested_size);
    buf->len = suggested_size;
}

void on_license_client_read(uv_stream_t *client, ssize_t nread,
                            const uv_buf_t *buf)
{
    KLOG_DEBUG("Receive client message. nread: %d.", (int32_t)nread);

    auto socket = (SocketManager *)client->data;
    if (nread < 0)
    {
        socket->del_connection((uv_tcp_t *)client);
        uv_close((uv_handle_t *)client, NULL);
        return;
    }

    if (nread > 0)
    {
        socket->read_connection((uv_tcp_t *)client, buf->base, nread);
    }

    if (buf->base != NULL)
    {
        free(buf->base);
    }
}

static void on_license_client_connected(uv_stream_t *server, int status)
{
    KLOG_DEBUG("connected status: %d.", status);

    if (status < 0)
    {
        KLOG_WARNING("New connection error: %s.", uv_strerror(status));
        return;
    }
    auto socket = (SocketManager *)server->data;
    auto client = (uv_tcp_t *)malloc(sizeof(uv_tcp_t));
    uv_tcp_init(server->loop, client);
    client->data = server->data;
    if (uv_accept(server, (uv_stream_t *)client) == 0)
    {
        socket->add_connection(client);
        uv_read_start((uv_stream_t *)client, alloc_client_buffer,
                      on_license_client_read);
    }
    else
    {
        uv_close((uv_handle_t *)client, NULL);
        KLOG_WARNING("Failed to accept a client connection. errno: %d.", errno);
        return;
    }
}

static void free_write_req(uv_write_t *req)
{
    write_req_t *wr = (write_req_t *)req;
    free(wr->buf.base);
    free(wr);
}

static void on_license_client_write(uv_write_t *req, int status)
{
    if (status > 0)
    {
        KLOG_WARNING("Write error: %s.", uv_strerror(status));
    }
    free_write_req(req);
}

void SocketManager::init()
{
    struct sockaddr_in addr;

    try
    {
        this->proxy_ = DBus::LicenseManagerProxyWrapper::create(this->dbus_connection_, LICENSE_MANAGER_OBJECT_PATH, LICENSE_MANAGER_DBUS_NAME);
        // TODO:
        // this->proxy_->dbusProxy()->set_default_timeout(LICENSE_DBUS_TIMEOUT_MSEC);
    }
    catch (const std::exception &e)
    {
        KLOG_WARNING("%s", e.what());
        return;
    }

    auto loop = uv_default_loop();
    uv_tcp_init(loop, &this->socket_server_);
    uv_ip4_addr("0.0.0.0", LICENSE_TCP_PORT, &addr);
    this->socket_server_.data = this;

    if (uv_tcp_bind(&this->socket_server_, (const struct sockaddr *)&addr, 0))
    {
        KLOG_WARNING("Failed to bind. addr: 0.0.0.0, port: %d errno: %d.",
                     LICENSE_TCP_PORT, errno);
        return;
    }

    if (uv_listen((uv_stream_t *)&this->socket_server_, LICENSE_SERVER_BACKLOG,
                  on_license_client_connected))
    {
        KLOG_WARNING("Failed to listen. errno: %d.", errno);
        return;
    }
}

bool SocketManager::add_connection(uv_tcp_t *client)
{
    uv_os_fd_t fd;

    KLOG_DEBUG("Client ip addr: %s", this->get_ip_addr(client).c_str());

    if (uv_fileno((uv_handle_t *)client, &fd))
    {
        KLOG_WARNING("Failed to get file descriptor.");
        return false;
    }
    auto iter = this->clients_.emplace(fd, ClientInfo{client, std::string()});
    if (!iter.second)
    {
        KLOG_WARNING("Exist same connection: %d.", fd);
        return false;
    }

    KLOG_DEBUG("Add file description %d for client %s.", fd, this->get_ip_addr(client).c_str());

    return true;
}

bool SocketManager::add_auth_connection(uv_tcp_t *client, const std::string &machine_code)
{
    uv_os_fd_t fd;
    if (uv_fileno((uv_handle_t *)client, &fd))
    {
        KLOG_WARNING("Failed to get file descriptor.");
        return false;
    }
    auto iter_client = this->clients_.find(fd);
    if (iter_client == this->clients_.end())
    {
        KLOG_WARNING("Not found client: %d.", fd);
        return false;
    }
    iter_client->second.machine_code = machine_code;

    auto iter_mc = this->machines_.emplace(machine_code, fd);

    if (!iter_mc.second && iter_mc.first->second != fd)
    {
        KLOG_WARNING("Exist a client that has the same machine code. : %d.", fd);
        return false;
    }
    return true;
}

bool SocketManager::del_connection(uv_tcp_t *client)
{
    uv_os_fd_t fd;
    if (uv_fileno((uv_handle_t *)client, &fd))
    {
        KLOG_WARNING("Failed to get file descriptor.");
        return false;
    }
    // 从数据库中删除
    // this->license_db_.del_cert_item(fd);

    auto iter = this->clients_.find(fd);
    if (iter == this->clients_.end())
    {
        KLOG_WARNING("Not found connection: %d.", fd);
        return false;
    }
    KLOG_DEBUG("Del client %s connection.", this->get_ip_addr(client).c_str());
    this->machines_.erase(iter->second.machine_code);
    this->clients_.erase(iter);
    return true;
}

void SocketManager::read_connection(uv_tcp_t *client, const char *data,
                                    size_t data_len)
{
    KLOG_DEBUG("Read client %s message. data len: %d.", this->get_ip_addr(client).c_str(), (int32_t)data_len);

    KiranLicenseMessage client_msg;
    if (!client_msg.ParseFromArray(data, data_len))
    {
        KLOG_WARNING("Failed parse data.");
        return;
    }

    KLOG_DEBUG("Receive message ID: %d.", client_msg.head().id());

    switch (client_msg.head().id())
    {
    case PROTO_ID_REGISTER_REQ:
        this->on_register(client, client_msg.body().register_req());
        break;
    case PROTO_ID_AUTH_REQ:
        this->on_auth(client, client_msg.body().auth_req());
        break;
    default:
        KLOG_WARNING("Unknown Message ID: %d.", client_msg.head().id());
        break;
    }
}

uv_os_fd_t SocketManager::get_fd(const std::string &machine_code)
{
    auto iter = this->machines_.find(machine_code);
    if (iter != this->machines_.end())
    {
        return iter->second;
    }
    return 0;
}

ClientInfo *SocketManager::get_client_info(const std::string &machine_code)
{
    auto fd = this->get_fd(machine_code);
    RETURN_VAL_IF_TRUE(fd == 0, NULL);
    auto iter = this->clients_.find(fd);
    if (iter != this->clients_.end())
    {
        return &(iter->second);
    }
    return NULL;
}

void SocketManager::on_register(uv_tcp_t *client, const KiranLicenseRegisterReq &register_req)
{
    KLOG_DEBUG("Receive client %s register. machine code: %s.",
               this->get_ip_addr(client).c_str(),
               register_req.machine_code().c_str());

    try
    {
        std::shared_ptr<UKeyDevice> used_ukey_device;
        auto ukey_group = UKeyManager::get_instance()->create_ukey_group(CLIENT_RY3_VENDOR_ID);

        if (!ukey_group)
        {
            KLOG_WARNING("Not found ukey group");
            this->send_error_msg(client, PROTO_ID_REGISTER_RSP, PROTO_ERROR_FAILED, LICENSE_ERROR2STR(LicenseErrorCode::LEC_UKEY_GROUP_UNINIT));
            return;
        }

        std::string activation_code;
        auto key = LicenseUtils::generate_key(LICENSE_OBJECT_OS_NAME);
        auto error_code = ukey_group->generate_license_code(register_req.machine_code(), key, used_ukey_device, activation_code);
        if (error_code != LicenseErrorCode::LEC_SUCCESS)
        {
            this->send_error_msg(client, PROTO_ID_REGISTER_RSP, PROTO_ERROR_FAILED, LICENSE_ERROR2STR(error_code));
            return;
        }

        this->message_.mutable_head()->set_id(PROTO_ID_REGISTER_RSP);
        this->message_.mutable_head()->set_error_code(PROTO_NO_ERROR);
        this->message_.mutable_body()->mutable_register_rsp()->set_machine_code(register_req.machine_code().c_str());
        this->message_.mutable_body()->mutable_register_rsp()->set_license_code(activation_code);
        this->send_license_msg(client, this->message_);
    }
    catch (const Glib::Error &e)
    {
        auto error = e.what();
        KLOG_WARNING("Client %s register error: %s.", this->get_ip_addr(client).c_str(), error.c_str());
        this->send_error_msg(client, PROTO_ID_REGISTER_RSP, PROTO_ERROR_FAILED, error);
        return;
    }
}

void SocketManager::on_auth(uv_tcp_t *client, const KiranLicenseAuthReq &auth_req)
{
    // std::string error;
    // bool auth_success = false;

    // this->message_.mutable_head()->set_id(PROTO_ID_AUTH_RSP);
    // this->message_.mutable_head()->set_error_code(PROTO_NO_ERROR);

    // KLOG_DEBUG("Receive client authorization. license code: %s, machine code: %s.",
    //           auth_req.license_code().c_str(),
    //           auth_req.machine_code().c_str());

    // do
    // {
    //     auto error_code = this->license_client_->check_with_mc(auth_req.machine_code(), auth_req.license_code());
    //     if (error_code != LicenseErrorCode::LEC_SUCCESS)
    //     {
    //         error = LICENSE_ERROR2STR(error_code);
    //         KLOG_DEBUG("Client %s auth error: %s.", this->get_ip_addr(client).c_str(), error.c_str());
    //         break;
    //     }

    //     // 判断是否有相同机器码的其他客户端连接
    //     auto client_info = this->get_client_info(auth_req.machine_code());
    //     if (client_info != NULL && client_info->socket != client)
    //     {
    //         error = _("Exist other clients that have the same machine code");
    //         break;
    //     }
    //     this->add_auth_connection(client, auth_req.machine_code());
    //     auth_success = true;

    // } while (0);

    // if (auth_success)
    // {
    //     this->message_.mutable_body()->mutable_auth_rsp()->set_result(PROTO_AUTH_SUCCESS);
    // }
    // else
    // {
    //     this->message_.mutable_head()->set_error_code(PROTO_ERROR_FAILED);
    //     this->message_.mutable_head()->set_error_desc(error);
    //     this->message_.mutable_body()->mutable_auth_rsp()->set_result(PROTO_AUTH_FAILED);
    // }

    // this->send_license_msg(client, this->message_);
}

void SocketManager::send_error_msg(uv_tcp_t *client,
                                   KiranLicenseProtoId id,
                                   KiranLicenseProtoError error_code,
                                   const std::string &error_desc)
{
    this->message_.mutable_head()->set_id(id);
    this->message_.mutable_head()->set_error_code(error_code);
    this->message_.mutable_head()->set_error_desc(error_desc);
    this->send_license_msg(client, this->message_);
}

void SocketManager::send_license_msg(uv_tcp_t *client, const KiranLicenseMessage &msg)
{
    auto msg_str = msg.SerializeAsString();
    write_req_t *req = (write_req_t *)malloc(sizeof(write_req_t));
    req->buf.base = (char *)malloc(msg_str.size());
    req->buf.len = msg_str.size();
    memcpy(req->buf.base, msg_str.c_str(), msg_str.size());

    uv_write((uv_write_t *)req, (uv_stream_t *)client, &req->buf, 1, on_license_client_write);
}

std::string SocketManager::get_ip_addr(uv_tcp_t *client)
{
    struct sockaddr_in peername;
    int peername_len = 0;
    char client_addr[IP4_ADDR_LEN + 1] = {0};
    std::string result;

    peername_len = sizeof(peername);
    if (uv_tcp_getpeername(client, (struct sockaddr *)&peername, &peername_len) < 0)
    {
        KLOG_DEBUG("Failed get peername.");
        return std::string();
    }

    if (uv_ip4_name(&peername, client_addr, IP4_ADDR_LEN) < 0)
    {
        KLOG_DEBUG("Failed get ip4 name.");
        return std::string();
    }
    return client_addr;
}

uv_os_fd_t SocketManager::get_fd(uv_tcp_t *client)
{
    uv_os_fd_t fd;
    if (uv_fileno((uv_handle_t *)client, &fd))
    {
        KLOG_WARNING("Failed to get file descriptor.");
        return 0;
    }
    return fd;
}

}  // namespace Kiran